from __future__ import annotations

import functools
import json
from concurrent.futures import ThreadPoolExecutor
from dataclasses import asdict, dataclass
from fnmatch import fnmatch
from itertools import zip_longest
from typing import TYPE_CHECKING, cast

from pdm.cli.commands.base import BaseCommand
from pdm.cli.utils import normalize_pattern
from pdm.models.requirements import strip_extras
from pdm.utils import normalize_name

if TYPE_CHECKING:
    from argparse import ArgumentParser, Namespace

    from unearth import PackageFinder

    from pdm.project.core import Project


@dataclass
class ListPackage:
    package: str
    groups: list[str]
    installed_version: str
    pinned_version: str
    latest_version: str = ""


@functools.lru_cache
def _find_first_diff(a: str, b: str) -> int:
    a_parts = a.split(".")
    b_parts = b.split(".")
    for i, (x, y) in enumerate(zip_longest(a_parts, b_parts)):
        if x != y:
            return (len(".".join(a_parts[:i])) + 1) if i > 0 else 0
    return 0


class Command(BaseCommand):
    """Check for outdated packages and list the latest versions on indexes."""

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "--json", action="store_const", const="json", dest="format", default="table", help="Output in JSON format"
        )
        parser.add_argument("--include-sub", action="store_true", help="Include sub-dependencies")
        parser.add_argument("patterns", nargs="*", help="The packages to check", type=normalize_pattern)

    @staticmethod
    def _match_pattern(name: str, patterns: list[str]) -> bool:
        return not patterns or any(fnmatch(name, p) for p in patterns)

    @staticmethod
    def _populate_latest_version(finder: PackageFinder, package: ListPackage) -> None:
        best = finder.find_best_match(package.package).best
        if best:
            package.latest_version = best.version or ""

    @staticmethod
    def _format_json(packages: list[ListPackage]) -> str:
        return json.dumps([asdict(package) for package in packages], indent=2)

    @staticmethod
    def _render_version(version: str, base_version: str) -> str:
        from packaging.version import InvalidVersion

        from pdm.utils import parse_version

        if not version or version == base_version:
            return version
        if not base_version:
            return f"[bold red]{version}[/]"

        try:
            parsed_version = parse_version(version)
            parsed_base_version = parse_version(base_version)
        except InvalidVersion:
            return version
        first_diff = _find_first_diff(version, base_version)
        head, tail = version[:first_diff], version[first_diff:]
        if parsed_version.major != parsed_base_version.major:
            return f"{head}[bold red]{tail}[/]"
        if parsed_version.minor != parsed_base_version.minor:
            return f"{head}[bold yellow]{tail}[/]"
        return f"{head}[bold green]{tail}[/]"

    def handle(self, project: Project, options: Namespace) -> None:
        environment = project.environment
        installed = environment.get_working_set()
        resolved = {strip_extras(k)[0]: v for k, v in project.get_locked_repository().candidates.items()}

        collected: list[ListPackage] = []
        project_dependencies: dict[str, list[str]] = {}
        for group, dependencies in project.all_dependencies.items():
            for dep in dependencies:
                project_dependencies.setdefault(cast(str, dep.key), []).append(group)

        for name, distribution in installed.items():
            if not self._match_pattern(name, options.patterns):
                continue
            if project.name and name == normalize_name(project.name):
                continue
            constrained_version = resolved.pop(name).version or "" if name in resolved else ""
            if not options.include_sub and name not in project_dependencies:
                continue
            groups = project_dependencies.get(name, [])
            collected.append(ListPackage(name, groups, distribution.version or "", constrained_version))

        for name, candidate in resolved.items():
            if not self._match_pattern(name, options.patterns):
                continue
            if candidate.req.marker and not candidate.req.marker.matches(environment.spec):
                continue
            if not options.include_sub and name not in project_dependencies:
                continue
            groups = project_dependencies.get(name, [])
            collected.append(ListPackage(name, groups, "", candidate.version or ""))

        with environment.get_finder() as finder, ThreadPoolExecutor() as executor:
            for package in collected:
                executor.submit(self._populate_latest_version, finder, package)

        collected = sorted(
            [p for p in collected if p.latest_version and p.latest_version != p.installed_version],
            key=lambda p: p.package,
        )
        if options.format == "json":
            print(self._format_json(collected))
        else:
            rows = [
                (
                    f"[bold]{package.package}[/]",
                    ", ".join(package.groups),
                    package.installed_version,
                    self._render_version(package.pinned_version, package.installed_version),
                    self._render_version(package.latest_version, package.installed_version),
                )
                for package in collected
            ]
            project.core.ui.display_columns(rows, header=["Package", "Groups", "Installed", "Pinned", "Latest"])



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING

from pdm.project.lockfile.base import (
    FLAG_CROSS_PLATFORM,
    FLAG_DIRECT_MINIMAL_VERSIONS,
    FLAG_INHERIT_METADATA,
    FLAG_STATIC_URLS,
    Lockfile,
)
from pdm.project.lockfile.pdmlock import PDMLock
from pdm.project.lockfile.pylock import PyLock

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


if TYPE_CHECKING:
    from pdm.project import Project

__all__ = [
    "FLAG_CROSS_PLATFORM",
    "FLAG_DIRECT_MINIMAL_VERSIONS",
    "FLAG_INHERIT_METADATA",
    "FLAG_STATIC_URLS",
    "Lockfile",
    "PDMLock",
    "PyLock",
    "load_lockfile",
]


def load_lockfile(project: Project, path: str | Path) -> Lockfile:
    """Load a lockfile from the given path."""

    default_lockfile = PyLock if project.config["lock.format"] == "pylock" else PDMLock

    try:
        with open(path, "rb") as f:
            data = tomllib.load(f)
    except OSError:
        return default_lockfile(path, ui=project.core.ui)
    else:
        klass: type[Lockfile]
        if data.get("metadata", {}).get("lock_version"):
            klass = PDMLock
        elif data.get("lock-version"):
            klass = PyLock
        else:  # pragma: no cover
            klass = default_lockfile
        lockfile = klass(path, ui=project.core.ui, parse=False)
        lockfile._data = data  # type: ignore[assignment]
        return lockfile



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from pdm.environments.base import BaseEnvironment
from pdm.models.in_process import get_sys_config_paths
from pdm.models.working_set import WorkingSet

if TYPE_CHECKING:
    from pdm.project import Project


class PythonEnvironment(BaseEnvironment):
    """A project environment that is directly derived from a Python interpreter"""

    def __init__(
        self,
        project: Project,
        *,
        python: str | None = None,
        prefix: str | None = None,
        extra_paths: list[str] | None = None,
    ) -> None:
        super().__init__(project, python=python)
        self.prefix = prefix
        self.extra_paths = extra_paths or []

    def get_paths(self, dist_name: str | None = None) -> dict[str, str]:
        is_venv = self.interpreter.get_venv() is not None
        if self.prefix is not None:
            replace_vars = {"base": self.prefix, "platbase": self.prefix}
            kind = "prefix"
        else:
            replace_vars = None
            kind = "user" if not is_venv and self.project.global_config["global_project.user_site"] else "default"
        paths = get_sys_config_paths(str(self.interpreter.executable), replace_vars, kind=kind)
        if is_venv:
            python_xy = f"python{self.interpreter.identifier}"
            paths["include"] = os.path.join(paths["data"], "include", "site", python_xy)
        elif not dist_name:
            dist_name = "UNKNOWN"
        if dist_name:
            paths["include"] = os.path.join(paths["include"], dist_name)
        paths["prefix"] = paths["data"]
        paths["headers"] = paths["include"]
        return paths

    @property
    def process_env(self) -> dict[str, str]:
        env = super().process_env
        venv = self.interpreter.get_venv()
        if venv is not None and self.prefix is None:
            env.update(venv.env_vars())
        return env

    def get_working_set(self) -> WorkingSet:
        scheme = self.get_paths()
        paths = [scheme["platlib"], scheme["purelib"]]
        venv = self.interpreter.get_venv()
        shared_paths = self.extra_paths[:]
        if venv is not None and venv.include_system_site_packages:
            shared_paths.extend(venv.base_paths)
        return WorkingSet(paths, shared_paths=list(dict.fromkeys(shared_paths)))



from __future__ import annotations

import collections
import dataclasses
import os
from functools import cached_property
from pathlib import Path
from typing import Any, Callable, ClassVar, Iterator, Mapping, MutableMapping, cast

import platformdirs
import rich.theme

from pdm import termui
from pdm._types import RepositoryConfig
from pdm.compat import tomllib
from pdm.exceptions import NoConfigError, PdmUsageError

REPOSITORY = "repository"
SOURCE = "pypi"
DEFAULT_REPOSITORIES = {
    "pypi": "https://upload.pypi.org/legacy/",
    "testpypi": "https://test.pypi.org/legacy/",
}

ui = termui.UI()


def choices(*choices: str) -> Callable[[str], str]:
    def converter(value: str) -> str:
        if value not in choices:
            raise ValueError(f"Value must be one of {choices}, got {value!r}")
        return value

    return converter


def load_config(file_path: Path) -> dict[str, Any]:
    """Load a nested TOML document into key-value pairs

    E.g. ["python"]["use_venv"] will be loaded as "python.use_venv" key.
    """

    def get_item(sub_data: Mapping[str, Any]) -> dict[str, Any]:
        result: dict[str, Any] = {}
        for k, v in sub_data.items():
            if k in (REPOSITORY, SOURCE):
                result.update((f"{k}.{sub_k}", sub_v) for sub_k, sub_v in v.items())
            elif isinstance(v, Mapping):
                result.update({f"{k}.{sub_k}": sub_v for sub_k, sub_v in get_item(v).items()})
            else:
                result.update({k: v})
        return result

    try:
        with file_path.open("rb") as fp:
            return get_item(tomllib.load(fp))
    except FileNotFoundError:
        return {}


def ensure_boolean(val: Any) -> bool:
    """Coerce a string value to a boolean value"""
    if not isinstance(val, str):
        return val

    return bool(val) and val.lower() not in ("false", "no", "0")


def split_by_comma(val: list[str] | str) -> list[str]:
    """Split a string value by comma"""
    if isinstance(val, str):
        return [v.strip() for v in val.split(",")]
    return val


DEFAULT_PYPI_INDEX = "https://pypi.org/simple"


@dataclasses.dataclass
class ConfigItem:
    """An item of configuration, with following attributes:

    Args:
        description (str): the config description
        default (Any): the default value, if given, will show in `pdm config`
        global_only (bool): not allowed to save in project config
        env_var (str|None): the env var name to take value from
        coerce (Callable): a function to coerce the value
        replace: (str|None): the deprecated name to replace
    """

    _NOT_SET = object()

    description: str
    default: Any = _NOT_SET
    global_only: bool = False
    env_var: str | None = None
    coerce: Callable = str
    replace: str | None = None

    def should_show(self) -> bool:
        return self.default is not self._NOT_SET


class Config(MutableMapping[str, str]):
    """A dict-like object for configuration key and values"""

    _config_map: ClassVar[dict[str, ConfigItem]] = {
        "cache_dir": ConfigItem(
            "The root directory of cached files",
            platformdirs.user_cache_dir("pdm"),
            True,
            env_var="PDM_CACHE_DIR",
        ),
        "log_dir": ConfigItem(
            "The root directory of log files",
            platformdirs.user_log_dir("pdm"),
            True,
            env_var="PDM_LOG_DIR",
        ),
        "check_update": ConfigItem(
            "Check if there is any newer version available",
            True,
            True,
            env_var="PDM_CHECK_UPDATE",
            coerce=ensure_boolean,
        ),
        "build_isolation": ConfigItem(
            "Isolate build environment from the project environment",
            True,
            False,
            "PDM_BUILD_ISOLATION",
            ensure_boolean,
        ),
        "request_timeout": ConfigItem(
            "The timeout for network requests in seconds", 15, True, "PDM_REQUEST_TIMEOUT", coerce=int
        ),
        "use_uv": ConfigItem(
            "Use uv for faster resolution and installation",
            False,
            False,
            "PDM_USE_UV",
            ensure_boolean,
        ),
        "lock.format": ConfigItem(
            "The format of the lock file, can be `pdm` or `pylock`",
            "pdm",
            env_var="PDM_LOCK_FORMAT",
            coerce=choices("pdm", "pylock"),
        ),
        "global_project.fallback": ConfigItem(
            "Use the global project implicitly if no local project is found",
            False,
            True,
            coerce=ensure_boolean,
        ),
        "global_project.fallback_verbose": ConfigItem(
            "If True show message when global project is used implicitly",
            True,
            True,
            coerce=ensure_boolean,
        ),
        "global_project.path": ConfigItem(
            "The path to the global project",
            platformdirs.user_config_path("pdm") / "global-project",
            True,
        ),
        "global_project.user_site": ConfigItem("Whether to install to user site", False, True, coerce=ensure_boolean),
        "strategy.update": ConfigItem("The default strategy for updating packages", "reuse", False),
        "strategy.save": ConfigItem("Specify how to save versions when a package is added", "minimum", False),
        "strategy.resolve_max_rounds": ConfigItem(
            "Specify the max rounds of resolution process",
            10000,
            env_var="PDM_RESOLVE_MAX_ROUNDS",
            coerce=int,
        ),
        "strategy.inherit_metadata": ConfigItem(
            "Inherit the groups and markers from parents for each package", True, coerce=ensure_boolean
        ),
        "install.parallel": ConfigItem(
            "Whether to perform installation and uninstallation in parallel",
            True,
            env_var="PDM_INSTALL_PARALLEL",
            coerce=ensure_boolean,
        ),
        "install.cache": ConfigItem(
            "Cache wheel installation and only put symlinks in the library root",
            False,
            coerce=ensure_boolean,
        ),
        "install.cache_method": ConfigItem(
            "Specify how to create links to the caches(`symlink/hardlink`)",
            "symlink",
        ),
        "python.providers": ConfigItem(
            "List of python provider names for findpython", default=[], coerce=split_by_comma
        ),
        "python.use_pyenv": ConfigItem("Use the pyenv interpreter", True, coerce=ensure_boolean),
        "python.use_venv": ConfigItem(
            "Use virtual environments when available", True, env_var="PDM_USE_VENV", coerce=ensure_boolean
        ),
        "python.use_python_version": ConfigItem(
            "Use .python-version file next to pyproject.toml to find python interpreters",
            True,
            env_var="PDM_USE_PYTHON_VERSION",
            coerce=ensure_boolean,
        ),
        "python.install_root": ConfigItem(
            "The root directory to install python interpreters",
            global_only=True,
            default=os.path.join(platformdirs.user_data_dir("pdm"), "python"),
        ),
        "pypi.url": ConfigItem(
            "The URL of PyPI mirror, defaults to https://pypi.org/simple",
            DEFAULT_PYPI_INDEX,
            env_var="PDM_PYPI_URL",
        ),
        "pypi.verify_ssl": ConfigItem(
            "Verify SSL certificate when query PyPI",
            True,
            env_var="PDM_PYPI_VERIFY_SSL",
            coerce=ensure_boolean,
        ),
        "pypi.username": ConfigItem("The username to access PyPI", env_var="PDM_PYPI_USERNAME"),
        "pypi.password": ConfigItem("The password to access PyPI", env_var="PDM_PYPI_PASSWORD"),
        "pypi.ca_certs": ConfigItem(
            "Path to a CA certificate bundle used for verifying the identity of the PyPI server", global_only=True
        ),
        "pypi.ignore_stored_index": ConfigItem(
            "Don't add the indexes from the config that is not listed in project",
            False,
            env_var="PDM_IGNORE_STORED_INDEX",
            coerce=ensure_boolean,
        ),
        "pypi.client_cert": ConfigItem("Path to client certificate file, or combined cert/key file", global_only=True),
        "pypi.client_key": ConfigItem("Path to client cert keyfile, if not in pypi.client_cert", global_only=True),
        "pypi.json_api": ConfigItem(
            "Consult PyPI's JSON API for package metadata",
            False,
            env_var="PDM_PYPI_JSON_API",
            coerce=ensure_boolean,
        ),
        "scripts.show_header": ConfigItem(
            "Display script name and help before running",
            default=False,
            env_var="PDM_SCRIPTS_SHOW_HEADER",
            coerce=ensure_boolean,
        ),
        "venv.location": ConfigItem(
            "Parent directory for virtualenvs",
            os.path.join(platformdirs.user_data_dir("pdm"), "venvs"),
            global_only=True,
        ),
        "venv.backend": ConfigItem(
            "Default backend to create virtualenv",
            default="virtualenv",
            env_var="PDM_VENV_BACKEND",
        ),
        "venv.in_project": ConfigItem(
            "Create virtualenv in `.venv` under project root",
            default=True,
            env_var="PDM_VENV_IN_PROJECT",
            coerce=ensure_boolean,
        ),
        "venv.prompt": ConfigItem(
            "Define a custom template to be displayed in the prompt when virtualenv is"
            "active. Variables `project_name` and `python_version` are available for"
            "formatting",
            default="{project_name}-{python_version}",
            env_var="PDM_VENV_PROMPT",
        ),
        "venv.with_pip": ConfigItem(
            "Install pip when creating a new venv",
            default=False,
            env_var="PDM_VENV_WITH_PIP",
            coerce=ensure_boolean,
        ),
    }
    _config_map.update(
        (f"theme.{k}", ConfigItem(f"Theme color for {k}", default=v, global_only=True))
        for k, v in termui.DEFAULT_THEME.items()
    )

    site: Config | None = None

    @classmethod
    def get_defaults(cls) -> dict[str, Any]:
        defaults = {k: v.default for k, v in cls._config_map.items() if v.should_show()}
        if cls.site is None:
            cls.site = Config(platformdirs.site_config_path("pdm") / "config.toml")
        defaults.update(cls.site.self_data)
        return defaults

    @cached_property
    def env_map(self) -> Mapping[str, Any]:
        return EnvMap(self._config_map)

    @classmethod
    def add_config(cls, name: str, item: ConfigItem) -> None:
        """Add or modify a config item"""
        cls._config_map[name] = item

    def __init__(self, config_file: Path, is_global: bool = False):
        self.is_global = is_global
        self.config_file = config_file.resolve()
        self.deprecated = {v.replace: k for k, v in self._config_map.items() if v.replace}
        self._file_data = load_config(self.config_file)
        self._data = collections.ChainMap(
            cast(MutableMapping[str, Any], self.env_map) if not is_global else {},
            self._file_data,
            self.get_defaults() if is_global else {},
        )

    def reload(self) -> None:
        """Reload the config from file"""
        new_data = load_config(self.config_file)
        self._file_data.clear()
        self._file_data.update(new_data)

    def load_theme(self) -> rich.theme.Theme:
        if not self.is_global:  # pragma: no cover
            raise PdmUsageError("Theme can only be loaded from global config")
        return rich.theme.Theme({k[6:]: v for k, v in self.items() if k.startswith("theme.")})

    @property
    def self_data(self) -> dict[str, Any]:
        return dict(self._file_data)

    def iter_sources(self) -> Iterator[RepositoryConfig]:
        for name, data in self._data.items():
            if name.startswith(f"{SOURCE}.") and name not in self._config_map and data:
                yield RepositoryConfig(**data, name=name[len(SOURCE) + 1 :], config_prefix=SOURCE)

    def _save_config(self) -> None:
        import tomlkit

        """Save the changed to config file."""
        self.config_file.parent.mkdir(parents=True, exist_ok=True)
        toml_data: dict[str, Any] = {}
        for key, value in self._file_data.items():
            *parts, last = key.split(".")
            temp = toml_data
            for part in parts:
                if part not in temp:
                    temp[part] = {}
                temp = temp[part]
            temp[last] = value

        with self.config_file.open("w", encoding="utf-8") as fp:
            tomlkit.dump(toml_data, fp)

    def __getitem__(self, key: str) -> Any:
        parts = key.split(".")
        if parts[0] in (REPOSITORY, SOURCE) and key not in self._config_map:
            if len(parts) < 2:
                raise PdmUsageError(f"Must specify a {parts[0]} name")
            repo = self.get_repository_config(parts[1], parts[0])
            if repo is None:
                raise KeyError(f"No {parts[0]} named {parts[1]}")
            if len(parts) >= 3:
                repo.populate_keyring_auth()
                return getattr(repo, parts[2])
            return repo

        if key not in self._config_map and key not in self.deprecated:
            raise NoConfigError(key)
        config_key = self.deprecated.get(key, key)
        config = self._config_map[config_key]

        if config_key in self._data:
            result = self._data[config_key]
        elif config.replace:
            result = self._data[config.replace]
        else:
            raise NoConfigError(key) from None
        return config.coerce(result)

    def __setitem__(self, key: str, value: Any) -> None:
        from pdm.models.auth import keyring

        parts = key.split(".")
        if parts[0] in (REPOSITORY, SOURCE) and key not in self._config_map:
            if len(parts) < 3:
                raise PdmUsageError(f"Set {parts[0]} config with [success]{parts[0]}.{{name}}.{{attr}}")
            index_key = ".".join(parts[:2])
            username = self._data.get(index_key, {}).get("username")  # type: ignore[call-overload]
            service = f"pdm-{index_key.replace('.', '-')}"
            if (
                parts[2] == "password"
                and self.is_global
                and username
                and keyring.save_auth_info(service, username, value)
            ):
                return
            if parts[2] == "verify_ssl":
                value = ensure_boolean(value)
            self._file_data.setdefault(index_key, {})[parts[2]] = value
            self._save_config()
            return

        if key not in self._config_map and key not in self.deprecated:
            raise NoConfigError(key)
        config_key = self.deprecated.get(key, key)
        config = self._config_map[config_key]
        if not self.is_global and config.global_only:
            raise ValueError(f"Config item '{key}' is not allowed to set in project config.")

        value = config.coerce(value)
        if key in self.env_map:
            ui.warn(f"the config is shadowed by env var '{config.env_var}', the value set won't take effect.")
        self._file_data[config_key] = value
        if config.replace:
            self._file_data.pop(config.replace, None)
        self._save_config()

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[str]:
        keys: set[str] = set()
        for key in self._data:
            if key in self.deprecated:
                key = self.deprecated[key]
            keys.add(key)
        return iter(keys)

    def __delitem__(self, key: str) -> None:
        from pdm.models.auth import keyring

        parts = key.split(".")
        if parts[0] in (REPOSITORY, SOURCE) and key not in self._config_map:
            if len(parts) < 2:
                raise PdmUsageError(f"Should specify the name of {parts[0]}")
            index_key = ".".join(parts[:2])
            username = self._data.get(index_key, {}).get("username")  # type: ignore[call-overload]
            service = f"pdm-{index_key.replace('.', '-')}"
            if len(parts) >= 3:
                index_key, attr = key.rsplit(".", 1)
                if attr == "password" and username:
                    keyring.delete_auth_info(service, username)
                self._file_data.get(index_key, {}).pop(attr, None)
            else:
                del self._file_data[key]
                if username:
                    keyring.delete_auth_info(service, username)
            self._save_config()
            return

        config_key = self.deprecated.get(key, key)
        config = self._config_map[config_key]
        self._file_data.pop(config_key, None)
        if config.replace:
            self._file_data.pop(config.replace, None)

        env_var = config.env_var
        if env_var is not None and env_var in os.environ:
            ui.warn(f"The config is shadowed by env var '{env_var}', set value won't take effect.")
        self._save_config()

    def get_repository_config(self, name_or_url: str, prefix: str) -> RepositoryConfig | None:
        """Get a repository or source by name or url."""
        repositories: dict[str, RepositoryConfig] = {}
        for k, v in self._data.items():
            if not k.startswith(f"{prefix}.") or k in self._config_map:
                continue
            key = k[len(prefix) + 1 :]
            repositories[key] = RepositoryConfig(**v, name=key, config_prefix=prefix)
        config: RepositoryConfig | None = None
        if "://" in name_or_url:
            config = next(
                (v for v in repositories.values() if v.url == name_or_url),
                RepositoryConfig(url=name_or_url, name="__unknown__", config_prefix=prefix),
            )
        else:
            config = repositories.get(name_or_url)

        if prefix == SOURCE or not self.is_global:
            return config

        if name_or_url in DEFAULT_REPOSITORIES:
            if config is None:
                return RepositoryConfig(url=DEFAULT_REPOSITORIES[name_or_url], name=name_or_url, config_prefix=prefix)
            config.passive_update(url=DEFAULT_REPOSITORIES[name_or_url])
        if name_or_url in DEFAULT_REPOSITORIES.values():
            name = next(k for k, v in DEFAULT_REPOSITORIES.items() if v == name_or_url)
            if config is None:
                return RepositoryConfig(
                    name=name,
                    config_prefix=prefix,
                    url=name_or_url,
                )
            config.passive_update(url=name_or_url)
            if config.name == "__unknown__":
                config.name = name
        return config


class EnvMap(Mapping[str, Any]):
    def __init__(self, config_items: Mapping[str, ConfigItem]) -> None:
        self._config_map = config_items

    def __repr__(self) -> str:
        return repr(dict(self))

    def __getitem__(self, k: str) -> Any:
        try:
            item = self._config_map[k]
            if item.env_var:
                return item.coerce(os.environ[item.env_var])
        except KeyError:
            pass
        raise KeyError(k)

    def __iter__(self) -> Iterator[str]:
        for key, item in self._config_map.items():
            if item.env_var and item.env_var in os.environ:
                yield key

    def __len__(self) -> int:
        return sum(1 for _ in self)



from __future__ import annotations

import contextlib
import hashlib
import json
import os
import stat
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Generic, Iterable, TypeVar

from packaging.utils import canonicalize_name, parse_wheel_filename

from pdm._types import CandidateInfo
from pdm.exceptions import PdmException
from pdm.models.cached_package import CachedPackage
from pdm.models.candidates import Candidate
from pdm.models.markers import EnvSpec
from pdm.termui import logger
from pdm.utils import atomic_open_for_write, create_tracked_tempdir, make_file_executable

if TYPE_CHECKING:
    from httpx import Client
    from unearth import Link


KT = TypeVar("KT")
VT = TypeVar("VT")


class JSONFileCache(Generic[KT, VT]):
    """A file cache that stores key-value pairs in a json file."""

    def __init__(self, cache_file: Path | str) -> None:
        self.cache_file = Path(cache_file)
        self._cache: dict[str, VT] = {}
        self._read_cache()

    def _read_cache(self) -> None:
        if not self.cache_file.exists():
            self._cache = {}
            return
        with self.cache_file.open() as fp:
            try:
                self._cache = json.load(fp)
            except json.JSONDecodeError:
                return

    def _write_cache(self) -> None:
        with self.cache_file.open("w") as fp:
            json.dump(self._cache, fp)

    def __contains__(self, obj: KT) -> bool:
        return self._get_key(obj) in self._cache

    @classmethod
    def _get_key(cls, obj: KT) -> str:
        return str(obj)

    def get(self, obj: KT) -> VT:
        key = self._get_key(obj)
        return self._cache[key]

    def set(self, obj: KT, value: VT) -> None:
        key = self._get_key(obj)
        self._cache[key] = value
        self._write_cache()


class CandidateInfoCache(JSONFileCache[Candidate, CandidateInfo]):
    """A cache manager that stores the
    candidate -> (dependencies, requires_python, summary) mapping.
    """

    @staticmethod
    def get_url_part(link: Link) -> str:
        import base64

        from pdm.utils import url_without_fragments

        url = url_without_fragments(link.split_auth()[1])
        return base64.urlsafe_b64encode(url.encode()).decode()

    @classmethod
    def _get_key(cls, obj: Candidate) -> str:
        # Name and version are set when dependencies are resolved,
        # so use them for cache key. Local directories won't be cached.
        if not obj.name or not obj.version:
            raise KeyError("The package is missing a name or version")
        extras = "[{}]".format(",".join(sorted(obj.req.extras))) if obj.req.extras else ""
        version = obj.version
        if obj.link is not None and not obj.req.is_named:
            version = cls.get_url_part(obj.link)
        return f"{obj.name}{extras}-{version}"


class HashCache:
    """Caches hashes of PyPI artifacts so we do not need to re-download them.

    Hashes are only cached when the URL appears to contain a hash in it and the
    cache key includes the hash value returned from the server). This ought to
    avoid issues where the location on the server changes.
    """

    FAVORITE_HASH = "sha256"
    STRONG_HASHES = ("sha256", "sha384", "sha512")

    def __init__(self, directory: Path | str) -> None:
        self.directory = Path(directory)

    def _read_from_link(self, link: Link, session: Client) -> Iterable[bytes]:
        if link.is_file:
            with open(link.file_path, "rb") as f:
                yield from f
        else:
            import httpx

            with session.stream("GET", link.normalized) as resp:
                try:
                    resp.raise_for_status()
                except httpx.HTTPStatusError as e:
                    raise PdmException(f"Failed to read from {link.redacted}: {e}") from e
                yield from resp.iter_bytes(chunk_size=8192)

    def _get_file_hash(self, link: Link, session: Client) -> str:
        h = hashlib.new(self.FAVORITE_HASH)
        logger.debug("Downloading link %s for calculating hash", link.redacted)
        for chunk in self._read_from_link(link, session):
            h.update(chunk)
        return ":".join([h.name, h.hexdigest()])

    def _should_cache(self, link: Link) -> bool:
        # For now, we only disable caching for local files.
        # We may add more when we know better about it.
        return not link.is_file

    def get_hash(self, link: Link, session: Client) -> str:
        # If there is no link hash (i.e., md5, sha256, etc.), we don't want
        # to store it.
        hash_value = self.get(link.url_without_fragment)
        if not hash_value:
            if link.hashes and link.hashes.keys() & self.STRONG_HASHES:
                logger.debug("Using hash in link for %s", link.redacted)
                hash_name = next(k for k in self.STRONG_HASHES if k in link.hashes)
                hash_value = f"{hash_name}:{link.hashes[hash_name]}"
            elif link.hash and link.hash_name in self.STRONG_HASHES:
                logger.debug("Using hash in link for %s", link.redacted)
                hash_value = f"{link.hash_name}:{link.hash}"
            else:
                hash_value = self._get_file_hash(link, session)
            if self._should_cache(link):
                self.set(link.url_without_fragment, hash_value)
        return hash_value

    def _get_path_for_key(self, key: str) -> Path:
        hashed = hashlib.sha224(key.encode("utf-8")).hexdigest()
        parts = (hashed[:2], hashed[2:4], hashed[4:6], hashed[6:8], hashed[8:])
        return self.directory.joinpath(*parts)

    def get(self, url: str) -> str | None:
        path = self._get_path_for_key(url)
        with contextlib.suppress(OSError, UnicodeError):
            return path.read_text("utf-8").strip()
        return None

    def set(self, url: str, hash: str) -> None:
        path = self._get_path_for_key(url)
        with contextlib.suppress(OSError, UnicodeError):
            path.parent.mkdir(parents=True, exist_ok=True)
            with atomic_open_for_write(path, encoding="utf-8") as fp:
                fp.write(hash)


class EmptyCandidateInfoCache(CandidateInfoCache):
    def get(self, obj: Candidate) -> CandidateInfo:
        raise KeyError

    def set(self, obj: Candidate, value: CandidateInfo) -> None:
        pass


class EmptyHashCache(HashCache):
    def get(self, url: str) -> str | None:
        return None

    def set(self, url: str, hash: str) -> None:
        pass


class WheelCache:
    """Caches wheels so we do not need to rebuild them.

    Wheels are only cached when the URL contains egg-info or is a VCS repository
    with an *immutable* revision. There might be more than one wheels built for
    one sdist, the one with most preferred tag will be returned.
    """

    def __init__(self, directory: Path | str) -> None:
        self.directory = Path(directory)
        self.ephemeral_directory = Path(create_tracked_tempdir(prefix="pdm-wheel-cache-"))

    def _get_candidates(self, path: Path) -> Iterable[Path]:
        if not path.exists():
            return
        for candidate in path.iterdir():
            if candidate.name.endswith(".whl"):
                yield candidate

    def _get_path_parts(self, link: Link, env_spec: EnvSpec) -> tuple[str, ...]:
        hash_key = {
            "url": link.url_without_fragment,
            # target env participates in the hash key to handle the some cases
            # where the sdist produces different wheels on different Pythons, and
            # the differences are not encoded in compatibility tags.
            "env_spec": env_spec.as_dict(),
        }
        if link.subdirectory:
            hash_key["subdirectory"] = link.subdirectory
        if link.hash and link.hash_name:
            hash_key[link.hash_name] = link.hash
        hashed = hashlib.sha224(
            json.dumps(hash_key, sort_keys=True, separators=(",", ":"), ensure_ascii=True).encode("utf-8")
        ).hexdigest()
        return (hashed[:2], hashed[2:4], hashed[4:6], hashed[6:])

    def get_path_for_link(self, link: Link, env_spec: EnvSpec) -> Path:
        parts = self._get_path_parts(link, env_spec)
        return self.directory.joinpath(*parts)

    def get_ephemeral_path_for_link(self, link: Link, env_spec: EnvSpec) -> Path:
        parts = self._get_path_parts(link, env_spec)
        return self.ephemeral_directory.joinpath(*parts)

    def get(self, link: Link, project_name: str | None, env_spec: EnvSpec) -> Path | None:
        if not project_name:
            return None
        canonical_name = canonicalize_name(project_name)

        candidate = self._get_from_path(self.get_path_for_link(link, env_spec), canonical_name, env_spec)
        if candidate is not None:
            return candidate
        return self._get_from_path(self.get_ephemeral_path_for_link(link, env_spec), canonical_name, env_spec)

    def _get_from_path(self, path: Path, canonical_name: str, env_spec: EnvSpec) -> Path | None:
        max_compatible_candidate: tuple[tuple[int, ...], Path | None] = ((-1, -1, -1, -1), None)
        for candidate in self._get_candidates(path):
            try:
                name, *_ = parse_wheel_filename(candidate.name)
            except ValueError:
                logger.debug("Ignoring invalid cached wheel %s", candidate.name)
                continue
            if canonical_name != canonicalize_name(name):
                logger.debug(
                    "Ignoring cached wheel %s with invalid project name %s, expected: %s",
                    candidate.name,
                    name,
                    canonical_name,
                )
                continue
            compat = env_spec.wheel_compatibility(candidate.name)
            if compat is None:
                continue
            if compat > max_compatible_candidate[0]:
                max_compatible_candidate = (compat, candidate)
        return max_compatible_candidate[1]


@lru_cache(maxsize=None)
def get_wheel_cache(directory: Path | str) -> WheelCache:
    return WheelCache(directory)


class PackageCache:
    def __init__(self, root: Path) -> None:
        self.root = root

    def cache_wheel(self, wheel: Path) -> CachedPackage:
        """Create a CachedPackage instance from a wheel file"""
        import zipfile

        dest = self.root.joinpath(f"{wheel.name}.cache")
        pkg = CachedPackage(dest, original_wheel=wheel)
        if dest.exists():
            return pkg
        dest.mkdir(parents=True, exist_ok=True)
        with pkg.lock():
            logger.info("Unpacking wheel into cached location %s", dest)
            with zipfile.ZipFile(wheel) as zf:
                try:
                    for item in zf.infolist():
                        target_path = zf.extract(item, dest)
                        mode = item.external_attr >> 16
                        is_executable = bool(mode and stat.S_ISREG(mode) and mode & 0o111)
                        if is_executable:
                            make_file_executable(target_path)
                except Exception:  # pragma: no cover
                    pkg.cleanup()  # cleanup on any error
                    raise
        return pkg

    def iter_packages(self) -> Iterable[CachedPackage]:
        for path in self.root.rglob("*.whl.cache"):
            p = CachedPackage(path)
            with p.lock():  # ensure the package is not being created
                pass
            yield p

    def cleanup(self) -> int:
        """Remove unused cached packages"""
        count = 0
        for pkg in self.iter_packages():
            if not any(os.path.exists(fn) for fn in pkg.referrers):
                pkg.cleanup()
                count += 1
        return count



from __future__ import annotations

import argparse
import dataclasses as dc
import json
import os
import re
import sys
from collections import OrderedDict
from fnmatch import fnmatch
from gettext import gettext as _
from json import dumps
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Iterable, Mapping, MutableMapping, cast, no_type_check

from packaging.specifiers import SpecifierSet
from resolvelib.structs import DirectedGraph
from rich.tree import Tree

from pdm import termui
from pdm.exceptions import PdmArgumentError, ProjectError
from pdm.models.specifiers import PySpecSet, get_specifier
from pdm.utils import comparable_version, is_path_relative_to, normalize_name, url_to_path

if TYPE_CHECKING:
    from argparse import Action, _ArgumentGroup

    from packaging.version import Version
    from resolvelib.resolvers import RequirementInformation, ResolutionImpossible

    from pdm.compat import Distribution
    from pdm.compat import importlib_metadata as im
    from pdm.models.candidates import Candidate
    from pdm.models.markers import EnvSpec
    from pdm.models.requirements import Requirement
    from pdm.project import Project


class PdmFormatter(argparse.RawDescriptionHelpFormatter):
    def start_section(self, heading: str | None) -> None:
        return super().start_section(termui.style(heading.title() if heading else "", style="warning"))

    def _format_usage(
        self,
        usage: str | None,
        actions: Iterable[Action],
        groups: Iterable[_ArgumentGroup],
        prefix: str | None,
    ) -> str:
        if prefix is None:
            prefix = "Usage: "
        result = super()._format_usage(usage, actions, groups, prefix)  # type: ignore[arg-type]
        # Remove continuous spaces
        result = re.sub(r" +", " ", result)
        if prefix:
            return result.replace(prefix, termui.style(prefix, style="warning"))
        return result

    def _format_action(self, action: Action) -> str:
        # determine the required width and the entry label
        help_position = min(self._action_max_length + 2, self._max_help_position)
        help_width = max(self._width - help_position, 11)
        action_width = help_position - self._current_indent - 2
        action_header = self._format_action_invocation(action)

        # no help; start on same line and add a final newline
        if not action.help:
            tup = "", self._current_indent, action_header
            action_header = "{0:>{1}}{2}\n".format(*tup)

        # short action name; start on the same line and pad two spaces
        elif len(action_header) <= action_width:
            tup = "", self._current_indent, action_header, action_width  # type: ignore[assignment]
            action_header = "{0:>{1}}{2:<{3}}  ".format(*tup)
            indent_first = 0

        # long action name; start on the next line
        else:
            tup = "", self._current_indent, action_header
            action_header = "{0:>{1}}{2}\n".format(*tup)
            indent_first = help_position

        # Special format for empty action_header
        # - No extra indent block
        # - Help info in the same indent level as subactions
        if not action_header.strip():
            action_header = ""
            help_position = self._current_indent
            indent_first = self._current_indent

        # collect the pieces of the action help
        parts = [termui.style(action_header, style="primary")]

        # if there was help for the action, add lines of help text
        if action.help:
            help_text = self._expand_help(action)
            help_lines = []
            for help_line in help_text.split("\n"):
                help_lines += self._split_lines(help_line, help_width)
            parts.append("{:>{}}{}\n".format("", indent_first, help_lines[0]))
            for line in help_lines[1:]:
                parts.append("{:>{}}{}\n".format("", help_position, line))

        # or add a newline if the description doesn't end with one
        elif not action_header.endswith("\n"):
            if action_header:
                parts.append("\n")

        # cancel out extra indent when action_header is empty
        if not action_header:
            self._dedent()
        # if there are any sub-actions, add their help as well
        for subaction in self._iter_indented_subactions(action):
            parts.append(self._format_action(subaction))
        # cancel out extra dedent when action_header is empty
        if not action_header:
            self._indent()

        # return a single string
        return self._join_parts(parts)


class ArgumentParser(argparse.ArgumentParser):
    """A standard argument parser but with title-cased help."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if sys.version_info >= (3, 14):
            kwargs["color"] = False
        kwargs["formatter_class"] = PdmFormatter
        kwargs["add_help"] = False
        super().__init__(*args, **kwargs)
        self.add_argument(
            "-h", "--help", action="help", default=argparse.SUPPRESS, help="Show this help message and exit."
        )
        self._optionals.title = "options"

    def parse_known_args(self, args: Any = None, namespace: Any = None) -> Any:
        args, argv = super().parse_known_args(args, namespace)
        if argv:
            msg = _("unrecognized arguments: %s")
            self.error(msg % " ".join(argv))
        return args, argv


def format_similar_command(root_command: str, commands: list[str], script_commands: list[str]) -> str:
    from difflib import get_close_matches

    similar_commands = get_close_matches(root_command, commands)
    similar_script_commands = get_close_matches(root_command, script_commands)
    commands_text = "\n".join([f"  - {cmd}" for cmd in similar_commands])
    script_commands_text = "\n".join([f"  - {cmd}" for cmd in similar_script_commands])
    message = f"[red]Command not found: {root_command}[/]"
    if commands_text:
        message += f"""
[green]Did you mean one of these commands?
{commands_text}[/]"""

    if script_commands_text:
        message += f"""
[yellow]{"Or" if commands_text else "Did you mean"} one of these script commands?
{script_commands_text}[/]"""
    return message


class ErrorArgumentParser(ArgumentParser):
    """A subclass of argparse.ArgumentParser that raises
    parsing error rather than exiting.

    This does the same as passing exit_on_error=False on Python 3.9+
    """

    def _parse_known_args(
        self, arg_strings: list[str], namespace: argparse.Namespace, *args: Any, **kwargs: Any
    ) -> tuple[argparse.Namespace, list[str]]:
        try:
            return super()._parse_known_args(arg_strings, namespace, *args, **kwargs)
        except argparse.ArgumentError as e:
            # We raise a dedicated error to avoid being caught by the caller
            raise PdmArgumentError(e) from e


@dc.dataclass(frozen=True)
class PackageNode:
    """An internal class for the convenience of dependency graph building."""

    name: str = dc.field(hash=True, compare=True)
    version: str | None = dc.field(compare=False)
    requirements: dict[str, Requirement] = dc.field(compare=False)

    def __repr__(self) -> str:
        return f"<Package {self.name}=={self.version}>"


def build_dependency_graph(
    working_set: Mapping[str, im.Distribution],
    env_spec: EnvSpec,
    selected: set[str] | None = None,
    include_sub: bool = True,
) -> DirectedGraph:
    """Build a dependency graph from locked result."""
    graph: DirectedGraph[PackageNode | None] = DirectedGraph()
    graph.add(None)  # sentinel parent of top nodes.
    node_with_extras: set[str] = set()

    def add_package(key: str, dist: Distribution | None) -> PackageNode:
        from pdm.models.requirements import filter_requirements_with_extras, strip_extras

        name, extras = strip_extras(key)
        extras = extras or ()
        reqs: dict[str, Requirement] = {}
        if dist:
            requirements = filter_requirements_with_extras(dist.requires or [], extras, include_default=True)
            for req in requirements:
                if not req.marker or req.marker.matches(env_spec):
                    reqs[req.identify()] = req
            version: str | None = dist.version
        else:
            version = None

        node = PackageNode(key, version, reqs)
        if node not in graph:
            if extras:
                node_with_extras.add(name)
            graph.add(node)
            if include_sub:
                for k in reqs:
                    child = add_package(k, working_set.get(strip_extras(k)[0]))
                    graph.connect(node, child)

        return node

    selected_map: dict[str, str] = {}
    for key in selected or ():
        name = key.split("[")[0]
        if len(key) >= len(selected_map.get(name, "")):
            # Ensure key with extras remains
            selected_map[name] = key
    for k, dist in working_set.items():
        if selected is not None:
            name = k.split("[")[0]
            if name not in selected_map:
                continue
            k = selected_map[name]
        add_package(k, dist)
    for node in list(graph):
        if node is not None and not list(graph.iter_parents(node)):
            # Top requirements
            if node.name in node_with_extras:
                # Already included in package[extra], no need to keep the top level
                # non-extra package.
                graph.remove(node)
            else:
                graph.connect(None, node)
    return graph


def specifier_from_requirement(requirement: Requirement) -> str:
    return str(requirement.specifier) or "Any"


def add_package_to_tree(
    root: Tree,
    graph: DirectedGraph,
    package: PackageNode,
    required: list[str],
    visited: frozenset[str] = frozenset(),
) -> None:
    """Format one package.

    :param graph: the dependency graph
    :param package: the package instance
    :param required: the version required by its parent
    :param visited: the visited package collection
    """
    version = (
        "[error][ not installed ][/]"
        if not package.version
        else f"[error]{package.version}[/]"
        if required
        and not any(s in ("Any", "This project") or SpecifierSet(s).contains(package.version) for s in required)
        else f"[warning]{package.version}[/]"
    )
    # escape deps with extras
    name = package.name.replace("[", r"\[") if "[" in package.name else package.name
    if package.name in visited:
        version = r"[error]\[circular][/]"
    req_str = f"[ required: {'&&'.join(required)} ]" if required else "[ Not required ]"
    node = root.add(f"[req]{name}[/] {version} {req_str}")
    if package.name in visited:
        return
    children = sorted(graph.iter_children(package), key=lambda p: p.name)
    for child in children:
        required = [specifier_from_requirement(package.requirements[child.name])]
        add_package_to_tree(node, graph, child, required, visited | {package.name})


def add_package_to_reverse_tree(
    root: Tree,
    graph: DirectedGraph,
    package: PackageNode,
    child: PackageNode | None = None,
    requires: str = "",
    visited: frozenset[str] = frozenset(),
) -> None:
    """Format one package for output reverse dependency graph."""
    version = "[error][ not installed ][/]" if not package.version else f"[warning]{package.version}[/]"
    if package.name in visited:
        version = r"[error]\[circular][/]"
    requires = (
        f"[ requires: [error]{requires}[/] ]"
        if requires not in ("Any", "")
        and child
        and child.version
        and not SpecifierSet(requires).contains(child.version)
        else ""
        if not requires
        else f"[ requires: {requires} ]"
    )
    name = package.name.replace("[", r"\[") if "[" in package.name else package.name
    node = root.add(f"[req]{name}[/] {version} {requires}")

    if package.name in visited:
        return
    parents: list[PackageNode] = sorted(filter(None, graph.iter_parents(package)), key=lambda p: p.name)
    for parent in parents:
        requires = specifier_from_requirement(parent.requirements[package.name])
        add_package_to_reverse_tree(node, graph, parent, package, requires, visited=visited | {package.name})
    return


def package_is_project(package: PackageNode, project: Project) -> bool:
    return project.is_distribution and package.name == normalize_name(project.name)


def _format_forward_dependency_graph(
    project: Project, graph: DirectedGraph[PackageNode | None], patterns: list[str]
) -> Tree:
    """Format dependency graph for output."""
    root = Tree("Dependencies", hide_root=True)

    def find_package_to_add(package: PackageNode) -> PackageNode | None:
        if not patterns:
            return package
        to_check = [package]
        while to_check:
            package = to_check.pop(0)
            if package and package_match_patterns(package, patterns):
                return package
            to_check.extend(graph.iter_children(package))
        return None

    all_dependencies = {
        r.identify(): r
        for deps in project.all_dependencies.values()
        for r in deps
        if not r.marker or r.marker.matches(project.environment.spec)
    }
    top_level_dependencies = {find_package_to_add(node) for node in graph.iter_children(None) if node}
    for package in sorted((node for node in top_level_dependencies if node), key=lambda p: p.name):
        required: set[str] = set()
        for parent in graph.iter_parents(package):
            if parent:
                r = specifier_from_requirement(parent.requirements[package.name])
            elif package.name in all_dependencies:
                r = specifier_from_requirement(all_dependencies[package.name])
            elif package_is_project(package, project):
                r = "This project"
            else:
                continue
            required.add(r)
        add_package_to_tree(root, graph, package, sorted(required))
    return root


def _format_reverse_dependency_graph(
    project: Project, graph: DirectedGraph[PackageNode | None], patterns: list[str]
) -> Tree:
    """Format reverse dependency graph for output."""

    def find_package_to_add(package: PackageNode) -> PackageNode | None:
        if not patterns:
            return package
        to_check = [package]
        while to_check:
            package = to_check.pop(0)
            if package and package_match_patterns(package, patterns):
                return package
            to_check.extend(graph.iter_parents(package))
        return None

    root = Tree("Dependencies", hide_root=True)
    leaf_nodes = {find_package_to_add(node) for node in graph if not list(graph.iter_children(node)) and node}
    for package in sorted((node for node in leaf_nodes if node), key=lambda p: p.name):
        if not package:
            continue
        add_package_to_reverse_tree(root, graph, package)
    return root


def build_forward_dependency_json_subtree(
    root: PackageNode,
    project: Project,
    graph: DirectedGraph[PackageNode | None],
    required_by: PackageNode | None = None,
    visited: frozenset[str] = frozenset(),
) -> dict:
    required: set[str] = set()
    all_dependencies = {
        r.identify(): r
        for deps in project.all_dependencies.values()
        for r in deps
        if not r.marker or r.marker.matches(project.environment.spec)
    }
    for parent in graph.iter_parents(root):
        if parent:
            r = specifier_from_requirement(parent.requirements[root.name])
        elif not package_is_project(root, project):
            requirements = required_by.requirements if required_by else all_dependencies
            if root.name in requirements:
                r = specifier_from_requirement(requirements[root.name])
            else:
                continue
        else:
            r = "This project"
        required.add(r)

    children = graph.iter_children(root) if root.name not in visited else []

    return OrderedDict(
        package=root.name,
        version=root.version,
        required="&&".join(sorted(required)),
        dependencies=sorted(
            (
                build_forward_dependency_json_subtree(p, project, graph, root, visited | {root.name})
                for p in children
                if p
            ),
            key=lambda d: d["package"],
        ),
    )


def build_reverse_dependency_json_subtree(
    root: PackageNode,
    project: Project,
    graph: DirectedGraph[PackageNode | None],
    requires: PackageNode | None = None,
    visited: frozenset[str] = frozenset(),
) -> dict:
    parents = graph.iter_parents(root) if root.name not in visited else []
    return OrderedDict(
        package=root.name,
        version=root.version,
        requires=specifier_from_requirement(root.requirements[requires.name]) if requires else None,
        dependents=sorted(
            (
                build_reverse_dependency_json_subtree(p, project, graph, root, visited | {root.name})
                for p in parents
                if p
            ),
            key=lambda d: d["package"],
        ),
    )


def package_match_patterns(package: PackageNode, patterns: list[str]) -> bool:
    return not patterns or any(fnmatch(package.name, pattern) for pattern in patterns)


def build_dependency_json_tree(
    project: Project, graph: DirectedGraph[PackageNode | None], reverse: bool, patterns: list[str]
) -> list[dict]:
    def find_package_to_add(package: PackageNode) -> PackageNode | None:
        if not patterns:
            return package
        to_check = [package]
        while to_check:
            package = to_check.pop(0)
            if package and package_match_patterns(package, patterns):
                return package
            if reverse:
                to_check.extend(graph.iter_parents(package))
            else:
                to_check.extend(graph.iter_children(package))
        return None

    top_level_packages: Iterable[PackageNode | None]
    if reverse:
        top_level_packages = filter(lambda n: not list(graph.iter_children(n)), graph)  # leaf nodes
        build_dependency_json_subtree: Callable = build_reverse_dependency_json_subtree
    else:
        top_level_packages = graph.iter_children(None)  # root nodes
        build_dependency_json_subtree = build_forward_dependency_json_subtree
    top_level_packages = {find_package_to_add(node) for node in top_level_packages if node}
    return [
        build_dependency_json_subtree(p, project, graph)
        for p in sorted((node for node in top_level_packages if node), key=lambda p: p.name)
    ]


def show_dependency_graph(
    project: Project,
    graph: DirectedGraph[PackageNode | None],
    reverse: bool = False,
    json: bool = False,
    patterns: list[str] | None = None,
) -> None:
    echo = project.core.ui.echo
    if patterns is None:
        patterns = []
    if json:
        echo(
            dumps(
                build_dependency_json_tree(project, graph, reverse, patterns),
                indent=2,
            )
        )
        return

    if reverse:
        tree = _format_reverse_dependency_graph(project, graph, patterns)
    else:
        tree = _format_forward_dependency_graph(project, graph, patterns)
    echo(tree)


def save_version_specifiers(
    requirements: Iterable[Requirement],
    resolved: dict[str, list[Candidate]],
    save_strategy: str,
) -> None:
    """Rewrite the version specifiers according to the resolved result and save strategy

    :param requirements: the requirements to be updated
    :param resolved: the resolved mapping
    :param save_strategy: compatible/safe_compatible/wildcard/exact
    """

    def candidate_version(candidates: list[Candidate]) -> Version | None:
        if len(candidates) > 1:
            return None
        c = candidates[0]
        assert c.version is not None
        return comparable_version(c.version)

    for r in requirements:
        name = r.identify()
        if r.is_named and not r.specifier and name in resolved:
            version = candidate_version(resolved[name])
            if version is None:
                continue
            if save_strategy == "exact":
                r.specifier = get_specifier(f"=={version}")
            elif save_strategy in ["compatible", "safe_compatible"]:
                if version.is_prerelease or version.is_devrelease:
                    r.specifier = get_specifier(f">={version},<{version.major + 1}")
                else:
                    if save_strategy == "compatible":
                        r.specifier = get_specifier(f"~={version.major}.{version.minor}")
                    else:
                        r.specifier = get_specifier(f"~={version}")
            elif save_strategy == "minimum":
                r.specifier = get_specifier(f">={version}")


def check_project_file(project: Project) -> None:
    """Check the existence of the project file and throws an error on failure."""
    if not project.pyproject.is_valid:
        raise ProjectError(
            "The pyproject.toml has not been initialized yet. You can do this by running [success]`pdm init`[/]."
        ) from None


def find_importable_files(project: Project) -> Iterable[tuple[str, Path]]:
    """Find all possible files that can be imported"""
    from pdm.formats import FORMATS

    for filename in ("Pipfile", "pyproject.toml", "requirements.in", "requirements.txt", "setup.py", "setup.cfg"):
        project_file = project.root / filename
        if not project_file.exists():
            continue
        for key, module in FORMATS.items():
            if module.check_fingerprint(project, project_file.as_posix()):
                yield key, project_file


@no_type_check
def set_env_in_reg(env_name: str, value: str) -> None:
    """Manipulate the WinReg, and add value to the
    environment variable if exists or create new.
    """
    import winreg

    value = os.path.normcase(value)

    with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as root:
        with winreg.OpenKey(root, "Environment", 0, winreg.KEY_ALL_ACCESS) as env_key:
            try:
                old_value, type_ = winreg.QueryValueEx(env_key, env_name)
                paths = [os.path.normcase(item) for item in old_value.split(os.pathsep)]
                if value in paths:
                    return
            except FileNotFoundError:
                paths, type_ = [], winreg.REG_EXPAND_SZ
            new_value = os.pathsep.join([value, *paths])
            winreg.SetValueEx(env_key, env_name, 0, type_, new_value)


def format_resolution_impossible(err: ResolutionImpossible) -> str:
    from pdm.resolver.python import PythonRequirement

    causes: list[RequirementInformation] = err.causes
    info_lines: set[str] = set()
    if not causes:
        return ""
    if all(isinstance(cause.requirement, PythonRequirement) for cause in causes):
        project_requires: PythonRequirement = next(cause.requirement for cause in causes if cause.parent is None)
        pyspec = cast(PySpecSet, project_requires.specifier)
        conflicting = [
            cause
            for cause in causes
            if cause.parent is not None and not cause.requirement.specifier.is_superset(pyspec)
        ]
        result = [
            "Unable to find a resolution because the following dependencies don't work "
            "on all Python versions in the range of the project's `requires-python`: "
            f"[success]{pyspec}[/]."
        ]
        for req, parent in conflicting:
            pyspec &= req.specifier
            info_lines.add(f"  {req.as_line()} (from {parent!r})")
        result.extend(sorted(info_lines))
        if pyspec.is_empty():
            result.append("Consider changing the version specifiers of the dependencies to be compatible")
        else:
            result.append(
                "A possible solution is to change the value of `requires-python` "
                f"in pyproject.toml to [success]{pyspec}[/]."
            )
        return "\n".join(result)

    if len(causes) == 1:
        return (
            "Unable to find a resolution for "
            f"[success]{causes[0].requirement.identify()}[/]\n"
            "Please make sure the package name is correct."
        )

    result = [
        "Unable to find a resolution for "
        f"[success]{causes[0].requirement.identify()}[/]\n"
        "because of the following conflicts:"
    ]
    for req, parent in causes:
        info_lines.add(f"  {req.as_line()} (from {parent if parent else 'project'})")
    result.extend(sorted(info_lines))
    result.append(
        "To fix this, you could loosen the dependency version constraints in "
        "pyproject.toml. See https://pdm-project.org/en/latest/usage/lockfile/"
        "#solve-the-locking-failure for more details."
    )
    return "\n".join(result)


def merge_dictionary(target: MutableMapping[Any, Any], input: Mapping[Any, Any], append_array: bool = True) -> None:
    """Merge the input dict with the target while preserving the existing values
    properly. This will update the target dictionary in place.
    List values will be extended, but only if the value is not already in the list.
    """
    for key, value in input.items():
        if key not in target:
            target[key] = value
        elif isinstance(target[key], dict):
            merge_dictionary(target[key], value, append_array=append_array)
        elif isinstance(target[key], list) and append_array:
            target[key].extend(x for x in value if x not in target[key])
            if hasattr(target[key], "multiline"):
                target[key].multiline(True)  # type: ignore[attr-defined]
        else:
            target[key] = value


def is_pipx_installation() -> bool:
    return sys.prefix.split(os.sep)[-3:-1] == ["pipx", "venvs"]


def is_homebrew_installation() -> bool:
    return "/libexec" in sys.prefix.replace("\\", "/")


def is_scoop_installation() -> bool:
    return os.name == "nt" and is_path_relative_to(sys.prefix, Path.home() / "scoop/apps/pdm")


def get_dist_location(dist: Distribution) -> str:
    direct_url = dist.read_text("direct_url.json")
    if not direct_url:
        return ""
    direct_url_data = json.loads(direct_url)
    url = cast(str, direct_url_data["url"])
    if url.startswith("file:"):
        path = url_to_path(url)
        editable = direct_url_data.get("dir_info", {}).get("editable", False)
        return f"{'-e ' if editable else ''}{path}"
    return ""


def get_pep582_path(project: Project) -> str:
    from pdm.compat import resources_open_binary

    script_dir = Path(__file__).parent.parent / "pep582"
    if script_dir.exists():
        return str(script_dir)

    script_dir = project.global_config.config_file.parent / "pep582"
    if script_dir.joinpath("sitecustomize.py").exists():
        return str(script_dir)
    script_dir.mkdir(parents=True, exist_ok=True)
    with resources_open_binary("pdm.pep582", "sitecustomize.py") as f:
        script_dir.joinpath("sitecustomize.py").write_bytes(f.read())
    return str(script_dir)


def use_venv(project: Project, name: str) -> None:
    from pdm.cli.commands.venv.utils import get_venv_with_name
    from pdm.environments import PythonEnvironment

    venv = get_venv_with_name(project, cast(str, name))
    project.core.ui.info(f"In virtual environment: [success]{venv.root}[/]", verbosity=termui.Verbosity.DETAIL)
    project.environment = PythonEnvironment(project, python=str(venv.interpreter))


def normalize_pattern(pattern: str) -> str:
    """Normalize a pattern to a valid name for a package."""
    return re.sub(r"[^A-Za-z0-9*?]+", "-", pattern).lower()



from __future__ import annotations

import argparse
from typing import TYPE_CHECKING

from pdm.cli.commands.base import BaseCommand
from pdm.cli.filters import GroupSelection
from pdm.cli.hooks import HookManager
from pdm.cli.options import (
    dry_run_option,
    frozen_lockfile_option,
    install_group,
    lockfile_option,
    override_option,
    skip_option,
    venv_option,
)
from pdm.exceptions import PdmUsageError, ProjectError
from pdm.utils import normalize_name

if TYPE_CHECKING:
    from typing import Collection

    from pdm.project import Project


class Command(BaseCommand):
    """Remove packages from pyproject.toml"""

    arguments = (
        *BaseCommand.arguments,
        install_group,
        dry_run_option,
        lockfile_option,
        override_option,
        frozen_lockfile_option,
        skip_option,
        venv_option,
    )

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-d",
            "--dev",
            default=False,
            action="store_true",
            help="Remove packages from dev dependencies",
        )
        parser.add_argument("-G", "--group", help="Specify the target dependency group to remove from")
        parser.add_argument(
            "--no-sync",
            dest="sync",
            default=True,
            action="store_false",
            help="Only write pyproject.toml and do not uninstall packages",
        )
        parser.add_argument("packages", nargs="+", help="Specify the packages to remove")

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        self.do_remove(
            project,
            selection=GroupSelection.from_options(project, options),
            sync=options.sync,
            packages=options.packages,
            no_editable=options.no_editable,
            no_self=options.no_self,
            dry_run=options.dry_run,
            fail_fast=options.fail_fast,
            hooks=HookManager(project, options.skip),
        )

    @staticmethod
    def do_remove(
        project: Project,
        selection: GroupSelection,
        sync: bool = True,
        packages: Collection[str] = (),
        no_editable: bool = False,
        no_self: bool = False,
        dry_run: bool = False,
        fail_fast: bool = False,
        hooks: HookManager | None = None,
    ) -> None:
        """Remove packages from working set and pyproject.toml"""

        from pdm.cli.actions import do_lock, do_sync
        from pdm.cli.utils import check_project_file
        from pdm.models.requirements import parse_requirement
        from pdm.utils import cd

        hooks = hooks or HookManager(project)
        check_project_file(project)
        if not packages:
            raise PdmUsageError("Must specify at least one package to remove.")
        group = selection.one()
        lock_groups = project.lockfile.groups
        project.pyproject.open_for_write()
        deps, setter = project.use_pyproject_dependencies(group, selection.dev or False)
        project.core.ui.echo(
            f"Removing {'[bold]global[/] ' if project.is_global else ''}packages from [primary]{group}[/] "
            f"{'dev-' if selection.dev else ''}dependencies: " + ", ".join(f"[req]{name}[/]" for name in packages)
        )
        tracked_names: set[str] = set()
        with cd(project.root):
            for name in packages:
                req = parse_requirement(name)
                matched_indexes = sorted((i for i, r in enumerate(deps) if req.matches(r)), reverse=True)
                if not matched_indexes:
                    raise ProjectError(f"[req]{name}[/] does not exist in [primary]{group}[/] dependencies.")
                for i in matched_indexes:
                    del deps[i]
                tracked_names.add(normalize_name(name))
        setter(deps)

        if not dry_run:
            project.pyproject.write()
        if lock_groups and group not in lock_groups:
            project.core.ui.warn(f"Group [success]{group}[/] isn't in lockfile, skipping lock.")
            return
        # It may remove the whole group, exclude it from lock groups first
        project_groups = project.iter_groups()
        if lock_groups is not None:
            lock_groups = [g for g in lock_groups if g in project_groups]
        do_lock(project, "reuse", dry_run=dry_run, tracked_names=tracked_names, hooks=hooks, groups=lock_groups)
        if sync:
            do_sync(
                project,
                selection=GroupSelection(project, default=False, groups=[group], exclude_non_existing=True),
                clean=True,
                no_editable=no_editable,
                no_self=no_self,
                dry_run=dry_run,
                fail_fast=fail_fast,
                hooks=hooks,
            )



from __future__ import annotations

import importlib.resources
import sys
from collections.abc import Iterator
from pathlib import Path
from typing import TYPE_CHECKING, ContextManager, Sequence, TypeVar

if TYPE_CHECKING:
    from typing import IO, Protocol

    class SupportsIdentify(Protocol):
        def identify(self) -> str: ...


if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

T = TypeVar("T", bound="SupportsIdentify")

if (
    not (sys.version_info[:2] == (3, 9) and sys.platform == "win32")
    # a bug on windows+py39 that zipfile path is not normalized
):

    def resources_open_binary(package: str, resource: str) -> IO[bytes]:
        return (importlib.resources.files(package) / resource).open("rb")

    def resources_read_text(package: str, resource: str, encoding: str = "utf-8", errors: str = "strict") -> str:
        with (importlib.resources.files(package) / resource).open("r", encoding=encoding, errors=errors) as f:
            return f.read()

    def resources_path(package: str, resource: str) -> ContextManager[Path]:
        return importlib.resources.as_file(importlib.resources.files(package) / resource)

else:
    resources_open_binary = importlib.resources.open_binary
    resources_read_text = importlib.resources.read_text
    resources_path = importlib.resources.path


if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata


Distribution = importlib_metadata.Distribution


class CompatibleSequence(Sequence[T]):  # pragma: no cover
    """A compatibility class for Sequence that also exposes `items()`, `keys()` and `values()` methods"""

    def __init__(self, data: Sequence[T]) -> None:
        self._data = data

    def __getitem__(self, index: str | slice | int) -> T | Sequence[T]:  # type: ignore[override]
        if isinstance(index, str):
            from pdm.utils import deprecation_warning

            deprecation_warning(
                "__getitem__ with a string key is deprecated on the requirements collection. It's not a mapping but a list",
                stacklevel=2,
            )
            for r in self._data:
                if r.identify() == index:
                    return r
            raise KeyError(index)
        return self._data[index]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[T]:
        return iter(self._data)

    def keys(self) -> Sequence[str]:
        from pdm.utils import deprecation_warning

        deprecation_warning(
            ".keys() is deprecated on the requirements collection, it's not a mapping but a list.", stacklevel=2
        )
        return [r.identify() for r in self._data]

    def values(self) -> Sequence[T]:
        from pdm.utils import deprecation_warning

        deprecation_warning(
            ".values() is deprecated on the requirements collection, it's not a mapping but a list.", stacklevel=2
        )
        return self._data

    def items(self) -> Iterator[tuple[str, T]]:
        from pdm.utils import deprecation_warning

        deprecation_warning(
            ".items() is deprecated on the requirements collection, it's not a mapping anymore.", stacklevel=2
        )
        for r in self._data:
            yield r.identify(), r


__all__ = ["CompatibleSequence", "Distribution", "importlib_metadata", "tomllib"]



from __future__ import annotations

from functools import cached_property
from typing import Iterable

from pdm.exceptions import PdmUsageError
from pdm.models.repositories.lock import LockedRepository
from pdm.project.lockfile.base import (
    FLAG_DIRECT_MINIMAL_VERSIONS,
    FLAG_INHERIT_METADATA,
    FLAG_STATIC_URLS,
    Compatibility,
    Lockfile,
)


class PyLock(Lockfile):
    SUPPORTED_FLAGS = frozenset([FLAG_DIRECT_MINIMAL_VERSIONS, FLAG_INHERIT_METADATA, FLAG_STATIC_URLS])

    @property
    def hash(self) -> tuple[str, str]:
        return next(iter(self._data.get("tool", {}).get("pdm", {}).get("hashes", {}).items()), ("", ""))

    def update_hash(self, hash_value: str, algo: str = "sha256") -> None:
        self._data.setdefault("tool", {}).setdefault("pdm", {}).setdefault("hashes", {})[algo] = hash_value

    @property
    def groups(self) -> list[str] | None:
        return [*self._data.get("dependency-groups", []), *self._data.get("extras", [])]

    @cached_property
    def default_strategies(self) -> set[str]:
        return {FLAG_INHERIT_METADATA, FLAG_STATIC_URLS}

    @property
    def strategy(self) -> set[str]:
        return set(self._data.get("tool", {}).get("pdm", {}).get("strategy", self.default_strategies))

    def apply_strategy_change(self, changes: Iterable[str]) -> set[str]:
        for change in changes:
            change = change.replace("-", "_").lower()
            if change.startswith("no_") and change[3:] != FLAG_DIRECT_MINIMAL_VERSIONS:
                raise PdmUsageError(f"Unsupported strategy change for pylock: {change}")
        return super().apply_strategy_change(changes)

    def format_lockfile(self, repository: LockedRepository, groups: Iterable[str] | None, strategy: set[str]) -> None:
        from pdm.formats.pylock import PyLockConverter

        converter = PyLockConverter(repository.environment.project, repository)
        data = converter.convert(groups)
        data["tool"]["pdm"]["strategy"] = sorted(strategy)
        self.set_data(data)

    def compatibility(self) -> Compatibility:  # pragma: no cover
        return Compatibility.SAME



from __future__ import annotations

import dataclasses as dc
import re
from typing import TYPE_CHECKING, Any, NamedTuple, TypeVar, Union

if TYPE_CHECKING:
    from typing import Protocol


def _normalize_pattern(pattern: str) -> str:
    return re.sub(r"[^A-Za-z0-9?*\[\]-]+", "-", pattern).lower()


@dc.dataclass
class RepositoryConfig:
    """Private dataclass to be subclassed"""

    config_prefix: str
    name: str

    url: str | None = None
    username: str | None = None
    password: str | None = dc.field(default=None, repr=False)
    verify_ssl: bool | None = None
    type: str | None = None
    ca_certs: str | None = None
    client_cert: str | None = None
    client_key: str | None = None
    include_packages: list[str] = dc.field(default_factory=list)
    exclude_packages: list[str] = dc.field(default_factory=list)

    def __post_init__(self) -> None:
        self.include_packages = [_normalize_pattern(p) for p in self.include_packages]
        self.exclude_packages = [_normalize_pattern(p) for p in self.exclude_packages]

    def populate_keyring_auth(self) -> None:
        if self.username is None or self.password is None:
            from pdm.models.auth import keyring

            service = f"pdm-{self.config_prefix}-{self.name}"
            auth = keyring.get_auth_info(service, self.username)
            if auth is not None:
                self.username, self.password = auth

    def passive_update(self, other: RepositoryConfig | None = None, **kwargs: Any) -> None:
        """An update method that prefers the existing value over the new one."""
        if other is not None:
            for k in other.__dataclass_fields__:
                v = getattr(other, k)
                if getattr(self, k) is None and v is not None:
                    setattr(self, k, v)
        for k, v in kwargs.items():
            if getattr(self, k) is None and v is not None:
                setattr(self, k, v)

    def __rich__(self) -> str:
        config_prefix = f"{self.config_prefix}.{self.name}." if self.name else f"{self.config_prefix}."
        lines: list[str] = []
        self.populate_keyring_auth()
        if self.url:
            lines.append(f"[primary]{config_prefix}url[/] = {self.url}")
        if self.username:
            lines.append(f"[primary]{config_prefix}username[/] = {self.username}")
        if self.password:
            lines.append(f"[primary]{config_prefix}password[/] = [i]<hidden>[/]")
        if self.verify_ssl is not None:
            lines.append(f"[primary]{config_prefix}verify_ssl[/] = {self.verify_ssl}")
        if self.type:
            lines.append(f"[primary]{config_prefix}type[/] = {self.type}")
        if self.ca_certs:
            lines.append(f"[primary]{config_prefix}ca_certs[/] = {self.ca_certs}")
        return "\n".join(lines)

    @property
    def url_with_credentials(self) -> HiddenText:
        from urllib.parse import urlsplit, urlunsplit

        from pdm.utils import expand_env_vars_in_auth, hide_url

        assert self.url is not None
        self.populate_keyring_auth()
        if not self.username or not self.password:
            return hide_url(expand_env_vars_in_auth(self.url))
        parsed = urlsplit(self.url)
        *_, netloc = parsed.netloc.rpartition("@")
        netloc = f"{self.username}:{self.password}@{netloc}"
        url = urlunsplit((parsed.scheme, netloc, parsed.path, parsed.query, parsed.fragment))
        return hide_url(url)


RequirementDict = Union[str, dict[str, Union[str, bool]]]
CandidateInfo = tuple[list[str], str, str]


class SearchResult(NamedTuple):
    name: str
    version: str
    summary: str


SearchResults = list[SearchResult]


if TYPE_CHECKING:
    from typing import Required, TypedDict

    class Comparable(Protocol):
        def __lt__(self, __other: Any) -> bool: ...

    SpinnerT = TypeVar("SpinnerT", bound="Spinner")

    class Spinner(Protocol):
        def update(self, text: str) -> None: ...

        def __enter__(self: SpinnerT) -> SpinnerT: ...

        def __exit__(self, *args: Any) -> None: ...

    class RichProtocol(Protocol):
        def __rich__(self) -> str: ...

    class FileHash(TypedDict, total=False):
        url: str
        hash: Required[str]
        file: str


class NotSetType:
    pass


NotSet = NotSetType()


@dc.dataclass(frozen=True)
class HiddenText:
    secret: str
    redacted: str = dc.field(compare=False, hash=False)

    def __str__(self) -> str:
        return self.redacted

    def __repr__(self) -> str:
        return repr(str(self))



import abc
import re

from pdm.project import Config, Project
from pdm.project.lockfile import FLAG_CROSS_PLATFORM
from pdm.termui import Verbosity
from pdm.utils import parse_version


class BaseFixer(abc.ABC):
    """Base class for fixers"""

    # A unique identifier for the fixer
    identifier: str
    # A boolean flag to indicate if the problem is breaking
    breaking: bool = False

    def __init__(self, project: Project) -> None:
        self.project = project

    def log(self, message: str, verbosity: Verbosity = Verbosity.DETAIL) -> None:
        self.project.core.ui.echo(message, verbosity=verbosity)

    @abc.abstractmethod
    def get_message(self) -> str:
        """Return a description of the problem"""

    @abc.abstractmethod
    def fix(self) -> None:
        """Perform the fix"""

    @abc.abstractmethod
    def check(self) -> bool:
        """Check if the problem exists"""


class ProjectConfigFixer(BaseFixer):
    """Fix the project config"""

    identifier = "project-config"

    def get_message(self) -> str:
        return (
            "[success]python.path[/] config needs to be moved to [info].pdm-python[/] and "
            "[info].pdm.toml[/] needs to be renamed to [info]pdm.toml[/]"
        )

    def _fix_gitignore(self) -> None:
        gitignore = self.project.root.joinpath(".gitignore")
        if not gitignore.exists():
            return
        content = gitignore.read_text("utf8")
        if ".pdm-python" not in content:
            content = re.sub(r"^\.pdm\.toml$", ".pdm-python", content, flags=re.M)
            gitignore.write_text(content, "utf8")

    def fix(self) -> None:
        old_file = self.project.root.joinpath(".pdm.toml")
        config = Config(old_file).self_data
        if not self.project.root.joinpath(".pdm-python").exists() and config.get("python.path"):
            self.log("Creating .pdm-python...", verbosity=Verbosity.DETAIL)
            self.project.root.joinpath(".pdm-python").write_text(config["python.path"])
        self.project.project_config  # access the project config to move the config items
        self.log("Moving .pdm.toml to pdm.toml...", verbosity=Verbosity.DETAIL)
        old_file.unlink()
        self.log("Fixing .gitignore...", verbosity=Verbosity.DETAIL)
        self._fix_gitignore()

    def check(self) -> bool:
        return self.project.root.joinpath(".pdm.toml").exists()


class PackageTypeFixer(BaseFixer):  # pragma: no cover
    identifier = "package-type"

    def get_message(self) -> str:
        package_type = self.project.pyproject.settings["package-type"]
        dist = str(package_type == "library").lower()
        return (
            rf'[success]package-type = "{package_type}"[/] has been renamed to '
            rf"[info]distribution = {dist}[/] under \[tool.pdm] table"
        )

    def check(self) -> bool:
        return "package-type" in self.project.pyproject.settings

    def fix(self) -> None:
        # Copy the project settings
        self.project.pyproject.open_for_write()
        settings = self.project.pyproject.settings.copy()

        # Pop the package type and convert it to a distribution type
        package_type = settings.pop("package-type")
        dist = package_type == "library"
        settings["distribution"] = dist

        # Update the project settings with the new distribution type
        self.project.pyproject._data["tool"].pop("pdm")
        self.project.pyproject.settings.update(settings)

        # Write the updated settings back to the project
        self.project.pyproject.write(False)


class LockStrategyFixer(BaseFixer):
    identifier = "deprecated-cross-platform"

    def get_message(self) -> str:
        return "Lock strategy [success]`cross_platform`[/] has been deprecated in favor of lock targets."

    def check(self) -> bool:
        from pdm.project.lockfile import PDMLock

        lockfile = self.project.lockfile
        if not isinstance(lockfile, PDMLock):  # pragma: no cover
            return False
        lockfile_version = lockfile.file_version
        if not lockfile_version or parse_version(lockfile_version) < parse_version("4.5.0"):
            return False
        return FLAG_CROSS_PLATFORM in lockfile.strategy

    def fix(self) -> None:
        strategies = self.project.lockfile.strategy - {FLAG_CROSS_PLATFORM}
        lockfile = self.project.lockfile.open_for_write()
        lockfile["metadata"]["strategy"] = sorted(strategies)
        self.project.lockfile.write(False)
        self.log("Lock strategy [success]`cross_platform` has been removed.", verbosity=Verbosity.DETAIL)



import argparse
import os

from pdm.cli.commands.base import verbose_option
from pdm.cli.commands.init import Command as InitCommand
from pdm.project.core import Project


class Command(InitCommand):
    """Create a new Python project at <project_path>"""

    supports_other_generator = False

    arguments = (verbose_option,)

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        super().add_arguments(parser)
        parser.add_argument("project_path", help="The path to create the new project")

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        new_project = project.core.create_project(
            options.project_path, global_config=options.config or os.getenv("PDM_CONFIG_FILE")
        )
        return super().handle(new_project, options)



from __future__ import annotations

import argparse
from typing import TYPE_CHECKING

from pdm.cli.commands.base import BaseCommand
from pdm.cli.options import venv_option
from pdm.exceptions import PdmUsageError
from pdm.models.candidates import Candidate
from pdm.models.project_info import ProjectInfo
from pdm.models.requirements import parse_requirement
from pdm.project import Project
from pdm.utils import normalize_name, parse_version

if TYPE_CHECKING:
    from unearth import Package


def filter_stable(package: Package) -> bool:
    assert package.version
    return not parse_version(package.version).is_prerelease


class Command(BaseCommand):
    """Show the package information"""

    metadata_keys = ("name", "version", "summary", "license", "platform", "keywords")

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        venv_option.add_to_parser(parser)
        parser.add_argument(
            "package",
            type=normalize_name,
            nargs=argparse.OPTIONAL,
            help="Specify the package name, or show this package if not given",
        )
        for option in self.metadata_keys:
            parser.add_argument(f"--{option}", action="store_true", help=f"Show {option}")

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        package = options.package
        if package:
            with project.environment.get_finder() as finder:
                best_match = finder.find_best_match(package, allow_prereleases=True)
            if not best_match.applicable:
                project.core.ui.warn(f"No match found for the package {package!r}")
                return
            latest = Candidate.from_installation_candidate(best_match.best, parse_requirement(package))
            latest_stable = next(filter(filter_stable, best_match.applicable), None)
            metadata = latest.prepare(project.environment).metadata
        else:
            if not project.is_distribution:
                raise PdmUsageError("This project is not a library")
            package = normalize_name(project.name)
            metadata = project.make_self_candidate(False).prepare(project.environment).prepare_metadata(True)
            latest_stable = None
        project_info = ProjectInfo.from_distribution(metadata)

        if any(getattr(options, key, None) for key in self.metadata_keys):
            for key in self.metadata_keys:
                if getattr(options, key, None):
                    project.core.ui.echo(getattr(project_info, key))
            return

        installed = project.environment.get_working_set().get(package)
        if latest_stable:
            project_info.latest_stable_version = str(latest_stable.version)
        if installed:
            project_info.installed_version = str(installed.version)
        project.core.ui.display_columns(list(project_info.generate_rows()))



from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Generator

from resolvelib import BaseReporter
from rich import get_console
from rich.live import Live
from rich.progress import BarColumn, Progress, SpinnerColumn, TaskProgressColumn, TimeElapsedColumn

from pdm.models.reporter import CandidateReporter, RichProgressReporter
from pdm.termui import SPINNER, UI, Verbosity, logger

if TYPE_CHECKING:
    from resolvelib.resolvers import Criterion, RequirementInformation, State
    from rich.console import Console, ConsoleOptions, RenderResult

    from pdm.models.candidates import Candidate
    from pdm.models.requirements import Requirement


def log_title(title: str) -> None:
    logger.info("=" * 8 + " " + title + " " + "=" * 8)


class LockReporter(BaseReporter):
    @contextmanager
    def make_candidate_reporter(self, candidate: Candidate) -> Generator[CandidateReporter]:
        yield CandidateReporter()

    def starting(self) -> Any:
        log_title("Start resolving requirements")

    def adding_requirement(self, requirement: Requirement, parent: Candidate) -> None:
        parent_line = f"(from {parent.name} {parent.version})" if parent else ""
        logger.info("  Adding requirement %s%s", requirement.as_line(), parent_line)

    def ending(self, state: State) -> Any:
        log_title("Resolution Result")
        if state.mapping:
            column_width = max(map(len, state.mapping.keys()))
            for k, can in state.mapping.items():
                if not can.req.is_named:
                    can_info = can.req.url
                    if can.req.is_vcs:
                        can_info = f"{can_info}@{can.get_revision()}"
                else:
                    can_info = can.version
                logger.info(f"  {k.rjust(column_width)} {can_info}")


class RichLockReporter(LockReporter):
    def __init__(self, requirements: list[Requirement], ui: UI) -> None:
        self.ui = ui
        self.console = get_console()
        self.requirements = requirements
        self.progress = Progress(
            "[progress.description]{task.description}",
            "[info]{task.fields[text]}",
            BarColumn(),
            TaskProgressColumn(),
            console=self.console,
        )
        self._spinner = Progress(
            SpinnerColumn(SPINNER, style="primary"),
            TimeElapsedColumn(),
            "[bold]{task.description}",
            "{task.fields[info]}",
            console=self.console,
        )
        self._spinner_task = self._spinner.add_task("Resolving dependencies", info="", total=1)
        self.live = Live(self)

    @contextmanager
    def make_candidate_reporter(self, candidate: Candidate) -> Generator[CandidateReporter]:
        task_id = self.progress.add_task(f"Resolving {candidate.format()}", text="", total=None)
        try:
            yield RichProgressReporter(self.progress, task_id)
        finally:
            self.progress.update(task_id, visible=False)
            if candidate._prepared:
                candidate._prepared.reporter = CandidateReporter()

    def update(self, description: str | None = None, info: str | None = None, completed: float | None = None) -> None:
        self._spinner.update(self._spinner_task, description=description, info=info, completed=completed)
        self.live.refresh()

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:  # pragma: no cover
        yield self._spinner
        yield self.progress

    def start(self) -> None:
        """Start the progress display."""
        if self.ui.verbosity < Verbosity.DETAIL:
            self.live.start(refresh=True)

    def stop(self) -> None:
        """Stop the progress display."""
        self.live.stop()
        if not self.console.is_interactive:  # pragma: no cover
            self.console.print()

    def __enter__(self) -> RichLockReporter:
        self.start()
        return self

    def __exit__(self, *args: Any) -> None:
        self.stop()

    def starting_round(self, index: int) -> None:
        log_title(f"Starting round {index}")

    def starting(self) -> None:
        """Called before the resolution actually starts."""
        log_title("Start resolving requirements")
        for req in self.requirements:
            logger.info("  " + req.as_line())

    def ending_round(self, index: int, state: State) -> None:
        """Called before each round of resolution ends.

        This is NOT called if the resolution ends at this round. Use `ending`
        if you want to report finalization. The index is zero-based.
        """
        resolved = len(state.mapping)
        to_resolve = len(state.criteria) - resolved
        self.update(info=f"[info]{resolved}[/] resolved, [info]{to_resolve}[/] to resolve")

    def rejecting_candidate(self, criterion: Criterion, candidate: Candidate) -> None:
        if not criterion.information:
            logger.info("Candidate rejected because it contains invalid metadata: %s", candidate)
            return
        *others, last = criterion.information
        logger.info(
            "Candidate rejected: %s because it introduces a new requirement %s"
            " that conflicts with other requirements:\n  %s",
            candidate,
            last.requirement.as_line(),  # type: ignore[attr-defined]
            "  \n".join(
                sorted({f"  {req.as_line()} (from {parent if parent else 'project'})" for req, parent in others})
            ),
        )

    def pinning(self, candidate: Candidate) -> None:
        """Called when adding a candidate to the potential solution."""
        logger.info("Adding new pin: %s %s", candidate.name, candidate.version)

    def resolving_conflicts(self, causes: list[RequirementInformation]) -> None:
        conflicts = sorted({f"  {req.as_line()} (from {parent if parent else 'project'})" for req, parent in causes})
        logger.info("Conflicts detected: \n%s", "\n".join(conflicts))



from __future__ import annotations

import argparse
from functools import cached_property
from typing import TYPE_CHECKING

from pdm.exceptions import PdmUsageError
from pdm.utils import normalize_name

if TYPE_CHECKING:
    from typing import Iterator, Sequence

    from pdm.project import Project


class GroupSelection:
    def __init__(
        self,
        project: Project,
        *,
        default: bool = True,
        dev: bool | None = None,
        groups: Sequence[str] = (),
        group: str | None = None,
        excluded_groups: Sequence[str] = (),
        exclude_non_existing: bool = False,
    ):
        self.project = project
        self.groups = groups
        self.group = group
        self.default = default
        self.dev = dev
        self.excluded_groups = excluded_groups
        self.exclude_non_existing = exclude_non_existing

    @classmethod
    def from_options(cls, project: Project, options: argparse.Namespace) -> GroupSelection:
        if getattr(options, "excluded_groups", None) and not options.groups and options.dev is None:
            options.groups = [":all"]
        if "group" in options:
            return cls(project, group=options.group, dev=options.dev)
        return cls(
            project,
            default=options.default,
            dev=options.dev,
            groups=options.groups,
            excluded_groups=options.excluded_groups,
        )

    def one(self) -> str:
        if self.group:
            return self.group
        if len(self.groups) == 1:
            return self.groups[0]
        return "dev" if self.dev else "default"

    @property
    def is_unset(self) -> bool:
        return self.default and self.dev is None and not self.groups

    def all(self) -> list[str] | None:
        project_groups = list(self.project.iter_groups())
        if self.is_unset:
            if self.project.lockfile.exists():
                groups = self.project.lockfile.groups
                if groups:
                    groups = [g for g in groups if g in project_groups]
                return groups
        return list(self)

    @cached_property
    def _translated_groups(self) -> list[str]:
        """Translate default, dev and groups containing ":all" into a list of groups"""
        locked_groups = self.project.lockfile.groups
        if self.is_unset:
            # Default case, return what is in the lock file
            project_groups = list(self.project.iter_groups())
            if locked_groups:
                return [g for g in locked_groups if g in project_groups]
        default, dev, groups = self.default, self.dev, self.groups
        if dev is None:  # --prod is not set, include dev-dependencies
            dev = True
        project = self.project
        optional_groups = {normalize_name(g) for g in project.pyproject.metadata.get("optional-dependencies", {})}
        dev_groups = set(project.pyproject.dev_dependencies)
        groups_set = {normalize_name(g) if g != ":all" else g for g in groups}
        if groups_set & dev_groups:
            if not dev:
                raise PdmUsageError("--prod is not allowed with dev groups and should be left")
        elif dev:
            groups_set.update(dev_groups)
            if self.exclude_non_existing and locked_groups:
                groups_set.intersection_update(locked_groups)
        if ":all" in groups:
            groups_set.discard(":all")
            groups_set.update(optional_groups)
        if default:
            groups_set.add("default")
        groups_set -= {normalize_name(g) for g in self.excluded_groups}

        invalid_groups = groups_set - {normalize_name(g) for g in project.iter_groups()}
        if invalid_groups:
            project.core.ui.echo(
                f"[d]Ignoring non-existing groups: [success]{', '.join(invalid_groups)}[/]",
                err=True,
            )
            groups_set -= invalid_groups
        # Sorts the result in ascending order instead of in random order
        # to make this function pure
        result = sorted(groups_set, key=lambda x: (x != "default", x))
        return result

    def validate(self) -> None:
        extra_groups = self.project.lockfile.compare_groups(self._translated_groups)
        if extra_groups:
            raise PdmUsageError(f"Requested groups not in lockfile: {','.join(extra_groups)}")

    def __iter__(self) -> Iterator[str]:
        return iter(self._translated_groups)

    def __contains__(self, group: str) -> bool:
        return group in self._translated_groups



from __future__ import annotations

import argparse
import sys

from pdm.cli.commands.base import BaseCommand
from pdm.compat import resources_read_text
from pdm.exceptions import PdmUsageError
from pdm.project import Project


class Command(BaseCommand):
    """Generate completion scripts for the given shell"""

    arguments = ()
    SUPPORTED_SHELLS = ("bash", "zsh", "fish", "powershell", "pwsh")

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "shell",
            nargs="?",
            help="The shell to generate the scripts for. If not given, PDM will properly guess from `SHELL` env var.",
        )

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        import shellingham

        shell = options.shell or shellingham.detect_shell()[0]
        if shell not in self.SUPPORTED_SHELLS:
            raise PdmUsageError(f"Unsupported shell: {shell}")
        suffix = "ps1" if shell in {"powershell", "pwsh"} else shell
        completion = resources_read_text("pdm.cli.completions", f"pdm.{suffix}")
        # Can't use rich print or otherwise the rich markups will be interpreted
        print(completion.replace("%{python_executable}", sys.executable))



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from pdm.cli.commands.base import BaseCommand
from pdm.cli.commands.venv.utils import iter_venvs
from pdm.cli.options import verbose_option

if TYPE_CHECKING:
    from argparse import Namespace

    from pdm.project import Project


class ListCommand(BaseCommand):
    """List all virtualenvs associated with this project"""

    arguments = (verbose_option,)

    def handle(self, project: Project, options: Namespace) -> None:
        project.core.ui.echo("Virtualenvs created with this project:\n")
        saved_python_root = Path(saved_python).parent.parent if (saved_python := project._saved_python) else None
        for ident, venv in iter_venvs(project):
            mark = "*" if saved_python_root and saved_python_root == venv.root else "-"
            project.core.ui.echo(f"{mark}  [success]{ident}[/]: {venv.root}")



from __future__ import annotations

import base64
import hashlib
from pathlib import Path
from typing import TYPE_CHECKING

from findpython import BaseProvider, PythonVersion

from pdm.exceptions import PdmUsageError
from pdm.models.venv import VirtualEnv

if TYPE_CHECKING:
    import sys
    from collections.abc import Iterable

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    from pdm.project import Project


def hash_path(path: str) -> str:
    """Generate a hash for the given path."""
    return base64.urlsafe_b64encode(hashlib.new("md5", path.encode(), usedforsecurity=False).digest()).decode()[:8]


def get_in_project_venv(root: Path) -> VirtualEnv | None:
    """Get the python interpreter path of venv-in-project"""
    for possible_dir in (".venv", "venv", "env"):
        venv = VirtualEnv.get(root / possible_dir)
        if venv is not None:
            return venv
    return None


def get_venv_prefix(project: Project) -> str:
    """Get the venv prefix for the project"""
    path = project.root
    name_hash = hash_path(path.as_posix())
    return f"{path.name}-{name_hash}-"


def iter_venvs(project: Project) -> Iterable[tuple[str, VirtualEnv]]:
    """Return an iterable of venv paths associated with the project"""
    in_project_venv = get_in_project_venv(project.root)
    if in_project_venv is not None:
        yield "in-project", in_project_venv
    venv_prefix = get_venv_prefix(project)
    venv_parent = Path(project.config["venv.location"])
    for path in venv_parent.glob(f"{venv_prefix}*"):
        ident = path.name[len(venv_prefix) :]
        venv = VirtualEnv.get(path)
        if venv is not None:
            yield ident, venv


def iter_central_venvs(project: Project) -> Iterable[tuple[str, Path]]:
    """Return an iterable of all managed venvs and their paths."""
    venv_parent = Path(project.config["venv.location"])
    for venv in venv_parent.glob("*"):
        ident = venv.name
        yield ident, venv


class VenvProvider(BaseProvider):
    """A Python provider for project venv pythons"""

    def __init__(self, project: Project) -> None:
        self.project = project

    @classmethod
    def create(cls) -> Self | None:
        return None

    def find_pythons(self) -> Iterable[PythonVersion]:
        for _, venv in iter_venvs(self.project):
            yield PythonVersion(venv.interpreter, _interpreter=venv.interpreter, keep_symlink=True)


def get_venv_with_name(project: Project, name: str) -> VirtualEnv:
    all_venvs = dict(iter_venvs(project))
    try:
        return all_venvs[name]
    except KeyError:
        raise PdmUsageError(
            f"No virtualenv with key '{name}' is found, must be one of {list(all_venvs)}.\n"
            "You can create one with 'pdm venv create'.",
        ) from None



"""
Special requirement and candidate classes to describe a requires-python constraint
"""

from __future__ import annotations

from typing import Iterable, Iterator, Mapping, cast

from pdm.models.candidates import Candidate
from pdm.models.requirements import NamedRequirement, Requirement
from pdm.models.specifiers import PySpecSet


class PythonCandidate(Candidate):
    def format(self) -> str:
        return f"[req]{self.name}[/][warning]{self.req.specifier!s}[/]"


class PythonRequirement(NamedRequirement):
    @classmethod
    def from_pyspec_set(cls, spec: PySpecSet) -> PythonRequirement:
        return cls(name="python", specifier=spec)

    def as_candidate(self) -> PythonCandidate:
        return PythonCandidate(self)


def find_python_matches(
    identifier: str,
    requirements: Mapping[str, Iterator[Requirement]],
) -> Iterable[Candidate]:
    """All requires-python except for the first one(must come from the project)
    must be superset of the first one.
    """
    python_reqs = cast(Iterator[PythonRequirement], iter(requirements[identifier]))
    project_req = next(python_reqs)
    python_specs = cast(Iterator[PySpecSet], (req.specifier for req in python_reqs))
    if all(spec.is_superset(project_req.specifier or "") for spec in python_specs):
        return [project_req.as_candidate()]
    else:
        # There is a conflict, no match is found.
        return []


def is_python_satisfied_by(requirement: Requirement, candidate: Candidate) -> bool:
    return cast(PySpecSet, requirement.specifier).is_superset(candidate.req.specifier)



from __future__ import annotations

import inspect
import os
from dataclasses import dataclass, replace
from pathlib import Path
from typing import cast

from pdm import termui
from pdm.models.candidates import Candidate
from pdm.models.markers import get_marker
from pdm.models.requirements import FileRequirement, Requirement, strip_extras
from pdm.models.specifiers import PySpecSet
from pdm.project.lockfile import FLAG_DIRECT_MINIMAL_VERSIONS, FLAG_INHERIT_METADATA
from pdm.resolver.base import Resolution, Resolver
from pdm.resolver.graph import merge_markers, populate_groups
from pdm.resolver.python import PythonRequirement
from pdm.resolver.reporters import LockReporter, RichLockReporter
from pdm.utils import normalize_name


@dataclass
class RLResolver(Resolver):
    def __post_init__(self) -> None:
        super().__post_init__()
        if self.locked_repository is None:
            self.locked_repository = self.project.get_locked_repository()
        supports_env_spec = "env_spec" in inspect.signature(self.project.get_provider).parameters
        if supports_env_spec:
            provider = self.project.get_provider(
                self.update_strategy,
                self.tracked_names,
                direct_minimal_versions=FLAG_DIRECT_MINIMAL_VERSIONS in self.strategies,
                env_spec=self.target,
                locked_repository=self.locked_repository,
            )
        else:  # pragma: no cover
            provider = self.project.get_provider(
                self.update_strategy,
                self.tracked_names,
                direct_minimal_versions=FLAG_DIRECT_MINIMAL_VERSIONS in self.strategies,
                ignore_compatibility=self.target.is_allow_all(),
            )
        if isinstance(self.reporter, LockReporter):
            provider.repository.reporter = self.reporter
        self.provider = provider

    def resolve(self) -> Resolution:
        from pdm.models.repositories import Package

        mapping = self._do_resolve()
        if self.project.enable_write_lockfile:  # type: ignore[has-type]
            if isinstance(self.reporter, RichLockReporter):
                self.reporter.update(info="Fetching hashes for resolved packages")
            self.provider.repository.fetch_hashes(mapping.values())
        if not (env_python := PySpecSet(self.target.requires_python)).is_superset(self.environment.python_requires):
            python_marker = get_marker(env_python.as_marker_string())
            for candidate in mapping.values():
                marker = candidate.req.marker or get_marker("")
                candidate.req = replace(candidate.req, marker=marker & python_marker)
        backend = self.project.backend
        packages: list[Package] = []
        for candidate in mapping.values():
            deps: list[str] = []
            for r in self.provider.fetched_dependencies[candidate.dep_key]:
                if isinstance(r, FileRequirement) and r.path:
                    try:
                        if r.path.is_absolute():
                            r.path = Path(os.path.normpath(r.path)).relative_to(os.path.normpath(self.project.root))
                    except ValueError:
                        pass
                    else:
                        r.url = backend.relative_path_to_url(r.path.as_posix())
                deps.append(r.as_line())
            packages.append(Package(candidate, deps, candidate.summary))
        return Resolution(packages, self.requested_groups)

    def _do_resolve(self) -> dict[str, Candidate]:
        from resolvelib import Resolver as _Resolver

        resolver_class = cast("type[_Resolver]", getattr(self.project.core, "resolver_class", _Resolver))
        resolver = resolver_class(self.provider, self.reporter)
        provider = self.provider
        repository = self.provider.repository
        target = self.target
        python_req = PythonRequirement.from_pyspec_set(PySpecSet(target.requires_python))
        requirements: list[Requirement] = [python_req, *self.requirements]
        max_rounds = self.project.config["strategy.resolve_max_rounds"]
        result = resolver.resolve(requirements, max_rounds)

        if repository.has_warnings:
            self.project.core.ui.info(
                "Use `-q/--quiet` to suppress these warnings, or ignore them per-package with "
                r"`ignore_package_warnings` config in \[tool.pdm] table.",
                verbosity=termui.Verbosity.NORMAL,
            )

        mapping = cast(dict[str, Candidate], result.mapping)
        mapping.pop("python", None)

        local_name = normalize_name(self.project.name) if self.project.is_distribution else None
        for key, candidate in list(mapping.items()):
            if key is None:
                continue
            # For source distribution whose name can only be determined after it is built,
            # the key in the resolution map and criteria should be updated.
            if key.startswith(":empty:"):
                new_key = provider.identify(candidate)
                mapping[new_key] = mapping.pop(key)
                result.criteria[new_key] = result.criteria.pop(key)  # type: ignore[attr-defined]

        if FLAG_INHERIT_METADATA in self.strategies:
            all_markers = merge_markers(result)
            populate_groups(result)
        else:
            all_markers = {}

        for key, candidate in list(mapping.items()):
            if key in all_markers:
                marker = all_markers[key]
                if marker.is_empty():
                    del mapping[key]
                    continue
                candidate.req = replace(candidate.req, marker=None if marker.is_any() else marker)

            if not self.keep_self and strip_extras(key)[0] == local_name:
                del mapping[key]

        return mapping



from __future__ import annotations

import itertools
from dataclasses import dataclass, field
from email.message import Message
from typing import TYPE_CHECKING, Iterator, cast

if TYPE_CHECKING:
    from pdm.compat import Distribution


@dataclass
class ProjectInfo:
    name: str
    version: str
    summary: str = ""
    author: str = ""
    email: str = ""
    license: str = ""
    requires_python: str = ""
    platform: str = ""
    keywords: str = ""
    homepage: str = ""
    project_urls: list[str] = field(default_factory=list)
    latest_stable_version: str = ""
    installed_version: str = ""

    @classmethod
    def from_distribution(cls, data: Distribution) -> ProjectInfo:
        metadata = cast(Message, data.metadata)
        keywords = metadata.get("Keywords", "").replace(",", ", ")
        platform = metadata.get("Platform", "").replace(",", ", ")

        if "Project-URL" in metadata:
            project_urls = {
                k.strip(): v.strip() for k, v in (row.split(",") for row in metadata.get_all("Project-URL", []))
            }
        else:
            project_urls = {}

        return cls(
            name=metadata.get("Name", ""),
            version=metadata.get("Version", ""),
            summary=metadata.get("Summary", ""),
            author=metadata.get("Author", ""),
            email=metadata.get("Author-email", ""),
            license=metadata.get("License", ""),
            requires_python=metadata.get("Requires-Python", ""),
            platform=platform,
            keywords=keywords,
            homepage=metadata.get("Home-page", ""),
            project_urls=[": ".join(parts) for parts in project_urls.items()],
        )

    def generate_rows(self) -> Iterator[tuple[str, str]]:
        yield "[primary]Name[/]:", self.name
        yield "[primary]Latest version[/]:", self.version
        if self.latest_stable_version:
            yield ("[primary]Latest stable version[/]:", self.latest_stable_version)
        if self.installed_version:
            yield ("[primary]Installed version[/]:", self.installed_version)
        yield "[primary]Summary[/]:", self.summary
        yield "[primary]Requires Python:", self.requires_python
        yield "[primary]Author[/]:", self.author
        yield "[primary]Author email[/]:", self.email
        yield "[primary]License[/]:", self.license
        yield "[primary]Homepage[/]:", self.homepage
        yield from itertools.zip_longest(("[primary]Project URLs[/]:",), self.project_urls, fillvalue="")
        yield "[primary]Platform[/]:", self.platform
        yield "[primary]Keywords[/]:", self.keywords



from __future__ import annotations

import argparse
import shlex
import subprocess
import sys
from typing import Any

from pdm import termui
from pdm.cli.actions import get_latest_pdm_version_from_pypi
from pdm.cli.commands.base import BaseCommand
from pdm.cli.options import verbose_option
from pdm.cli.utils import PackageNode, build_dependency_graph
from pdm.compat import Distribution
from pdm.environments import BareEnvironment
from pdm.models.markers import EnvSpec
from pdm.models.working_set import WorkingSet
from pdm.project import Project
from pdm.utils import is_in_zipapp, normalize_name, parse_version

PDM_REPO = "https://github.com/pdm-project/pdm"


def list_distributions(plugin_only: bool = False) -> list[Distribution]:
    result: list[Distribution] = []
    working_set = WorkingSet()
    for dist in working_set.values():
        if not plugin_only or any(ep.group in ("pdm", "pdm.plugin") for ep in dist.entry_points):
            result.append(dist)
    return sorted(result, key=lambda d: d.metadata.get("Name", "UNKNOWN"))


def run_pip(project: Project, args: list[str]) -> subprocess.CompletedProcess[str]:
    if project.config["use_uv"]:
        if "--upgrade-strategy" in args:
            # uv doesn't support this option
            args[(i := args.index("--upgrade-strategy")) : i + 2] = []
        run_args = [*project.core.uv_cmd, "pip", *args, "--python", sys.executable]
    else:
        env = BareEnvironment(project)
        project.environment = env
        run_args = [*env.pip_command, *args]
    project.core.ui.echo(f"Running pip command: {run_args}", verbosity=termui.Verbosity.DETAIL)

    result = subprocess.run(
        run_args,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        check=True,
        text=True,
    )
    project.core.ui.echo(
        f"Run pip returns status {result.returncode}: {result.stdout}", verbosity=termui.Verbosity.DEBUG
    )
    return result


class Command(BaseCommand):
    """Manage the PDM program itself (previously known as plugin)"""

    arguments = (verbose_option,)
    name = "self"

    @classmethod
    def register_to(
        cls,
        subparsers: argparse._SubParsersAction,
        name: str | None = None,
        **kwargs: Any,
    ) -> None:
        return super().register_to(subparsers, name, aliases=["plugin"], **kwargs)

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        subparsers = parser.add_subparsers(title="commands", metavar="")
        ListCommand.register_to(subparsers)
        if not is_in_zipapp():
            AddCommand.register_to(subparsers)
            RemoveCommand.register_to(subparsers)
            UpdateCommand.register_to(subparsers)
        parser.set_defaults(search_parent=False)
        self.parser = parser

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        self.parser.print_help()


class ListCommand(BaseCommand):
    """List all packages installed with PDM"""

    arguments = (verbose_option,)
    name = "list"

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("--plugins", action="store_true", help="List plugins only")

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        distributions = list_distributions(plugin_only=options.plugins)
        echo = project.core.ui.echo
        if not distributions:
            # This should not happen when plugin_only is False
            echo("No plugin is installed with PDM", err=True)
            sys.exit(1)
        echo("Installed packages:", err=True)
        rows = []
        for dist in distributions:
            rows.append(
                (
                    f"[success]{dist.metadata.get('Name')}[/]",
                    f"[warning]{dist.metadata.get('Version')}[/]",
                    dist.metadata.get("Summary", ""),
                ),
            )
        project.core.ui.display_columns(rows)


class AddCommand(BaseCommand):
    """Install packages to the PDM's environment"""

    arguments = (verbose_option,)
    name = "add"

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--pip-args",
            help="Arguments that will be passed to pip install",
            default="",
        )
        parser.add_argument(
            "packages",
            nargs="+",
            help="Specify one or many package names, each package can have a version specifier",
        )

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        pip_args = ["install", *shlex.split(options.pip_args), *options.packages]

        try:
            with project.core.ui.open_spinner(f"Installing packages: {options.packages}"):
                run_pip(project, pip_args)
        except subprocess.CalledProcessError as e:
            project.core.ui.echo("[error]Installation failed:[/]\n" + e.output, err=True)
            sys.exit(1)
        else:
            project.core.ui.echo("[success]Installation succeeds.[/]")


class RemoveCommand(BaseCommand):
    """Remove packages from PDM's environment"""

    arguments = (verbose_option,)
    name = "remove"

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--pip-args",
            help="Arguments that will be passed to pip uninstall",
            default="",
        )
        parser.add_argument("-y", "--yes", action="store_true", help="Answer yes on the question")
        parser.add_argument("packages", nargs="+", help="Specify one or many package names")

    def _resolve_dependencies_to_remove(self, packages: list[str]) -> list[str]:
        """Perform a BFS to find all unneeded dependencies"""
        result: set[str] = set()
        to_resolve = list(packages)

        ws = WorkingSet()
        graph = build_dependency_graph(ws, env_spec=EnvSpec.current())
        while to_resolve:
            temp: list[PackageNode] = []
            for name in to_resolve:
                key = normalize_name(name)
                if key in ws:
                    result.add(key)
                package = PackageNode(key, "0.0.0", {})
                if package not in graph:
                    continue
                for dep in graph.iter_children(package):
                    temp.append(dep)
                graph.remove(package)

            to_resolve.clear()
            for dep in temp:
                if not any(graph.iter_parents(dep)) and dep.name != "pdm":
                    to_resolve.append(dep.name)

        return sorted(result)

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        packages_to_remove = self._resolve_dependencies_to_remove(options.packages)
        if not packages_to_remove:
            project.core.ui.echo("No package to remove.", err=True)
            sys.exit(1)
        if not (options.yes or termui.confirm(f"Will remove: {packages_to_remove}, continue?", default=True)):
            return
        pip_args = ["uninstall", "-y", *shlex.split(options.pip_args), *packages_to_remove]

        try:
            with project.core.ui.open_spinner(f"Uninstalling packages: [success]{', '.join(options.packages)}[/]"):
                run_pip(project, pip_args)
        except subprocess.CalledProcessError as e:
            project.core.ui.echo("[error]Uninstallation failed:[/]\n" + e.output, err=True)
            sys.exit(1)
        else:
            project.core.ui.echo("[success]Uninstallation succeeds.[/]")


class UpdateCommand(BaseCommand):
    """Update PDM itself"""

    arguments = (verbose_option,)
    name = "update"

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--head",
            action="store_true",
            help="Update to the latest commit on the main branch",
        )
        parser.add_argument(
            "--pre",
            help="Update to the latest prerelease version",
            action="store_true",
        )
        parser.add_argument(
            "--no-frozen-deps",
            action="store_false",
            dest="frozen_deps",
            default=True,
            help="Do not install frozen dependency versions",
        )
        parser.add_argument(
            "--pip-args",
            help="Additional arguments that will be passed to pip install",
            default="",
        )

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        from pdm.__version__ import __version__, read_version

        locked = "[locked]" if options.frozen_deps else ""

        if options.head:
            package = f"pdm{locked} @ git+{PDM_REPO}@main"
            version: str | None = "HEAD"
        else:
            version = get_latest_pdm_version_from_pypi(project, options.pre)
            assert version is not None, "No version found"
            if parse_version(__version__) >= parse_version(version):
                project.core.ui.echo(f"Already up-to-date: [primary]{__version__}[/]")
                return
            package = f"pdm{locked}=={version}"
        pip_args = ["install", "--upgrade", "--upgrade-strategy", "eager", *shlex.split(options.pip_args), package]
        try:
            with project.core.ui.open_spinner(f"Updating pdm to version [primary]{version}[/]"):
                run_pip(project, pip_args)
        except subprocess.CalledProcessError as e:
            project.core.ui.echo(
                f"[error]Installing version [primary]{version}[/] failed:[/]\n" + e.output,
                err=True,
            )
            sys.exit(1)
        else:
            project.core.ui.echo(f"[success]Successfully installed version [primary]{version}[/][/]")
            project.core.ui.echo(f"See what's new in this version: [link]{PDM_REPO}/releases/tag/{version}[/]")
            # Update the version value to avoid check update print wrong message
            project.core.version = read_version()



from __future__ import annotations

import dataclasses
import json
from functools import cached_property
from itertools import chain
from typing import Collection, Iterable

from pdm import termui
from pdm.compat import Distribution
from pdm.environments import BaseEnvironment
from pdm.exceptions import BuildError
from pdm.installers.manager import InstallManager
from pdm.models.candidates import Candidate
from pdm.models.repositories import Package
from pdm.models.requirements import FileRequirement, Requirement, parse_requirement, strip_extras
from pdm.utils import is_editable, normalize_name


def editables_candidate(environment: BaseEnvironment) -> Candidate | None:
    """Return a candidate for `editables` package"""
    with environment.get_finder() as finder:
        best = finder.find_best_match("editables").best
    return None if best is None else Candidate.from_installation_candidate(best, parse_requirement("editables"))


class BaseSynchronizer:
    """Synchronize the working set with given installation candidates

    :param candidates: a dict of candidates to be installed
    :param environment: the environment associated with the project
    :param clean: clean unneeded packages
    :param dry_run: only prints summary but do not install or uninstall
    :param retry_times: retry times when installation failed
    :param install_self: whether to install self project
    :param no_editable: if True, override all editable installations,
        if a list, override editables with the given names
    :param use_install_cache: whether to use install cache
    :param reinstall: whether to reinstall all packages
    :param only_keep: If true, only keep the selected candidates
    :param fail_fast: If true, stop the installation on first error
    """

    SEQUENTIAL_PACKAGES = ("pip", "setuptools", "wheel")

    def __init__(
        self,
        environment: BaseEnvironment,
        candidates: dict[str, Candidate] | None = None,
        clean: bool = False,
        dry_run: bool = False,
        retry_times: int = 1,
        install_self: bool = False,
        no_editable: bool | Collection[str] = False,
        reinstall: bool = False,
        only_keep: bool = False,
        fail_fast: bool = False,
        use_install_cache: bool | None = None,
        packages: Iterable[Package] = (),
        requirements: Iterable[Requirement] | None = None,
    ) -> None:
        if candidates:  # pragma: no cover
            self.requested_candidates = candidates
        else:
            self.requested_candidates = {entry.candidate.identify(): entry.candidate for entry in packages}
        self.environment = environment
        self.clean = clean
        self.dry_run = dry_run
        self.retry_times = retry_times
        self.no_editable = no_editable
        self.install_self = install_self
        if use_install_cache is None:
            use_install_cache = bool(environment.project.config["install.cache"])
        self.use_install_cache = use_install_cache
        self.reinstall = reinstall
        self.only_keep = only_keep
        self.parallel = environment.project.config["install.parallel"]
        self.fail_fast = fail_fast

        self.working_set = environment.get_working_set()
        self.ui = environment.project.core.ui
        self._manager: InstallManager | None = None
        self.packages = packages
        self.requirements = requirements

    @cached_property
    def self_candidate(self) -> Candidate:
        """Return the candidate for self project"""
        return self.environment.project.make_self_candidate(not self.no_editable)

    @cached_property
    def candidates(self) -> dict[str, Candidate]:
        """Return the candidates to be installed"""
        candidates = self.requested_candidates.copy()
        requested = {
            req.identify()
            for req in (self.requirements or chain.from_iterable(self.environment.project.all_dependencies.values()))
        }
        if isinstance(self.no_editable, Collection):
            keys = self.no_editable
        elif self.no_editable:
            keys = candidates.keys()
        else:
            keys = []
            if self.should_install_editables():
                # Install `editables` as well as required by self project
                editables = editables_candidate(self.environment)
                if editables is not None:
                    candidates["editables"] = editables
        for key in keys:
            if key in candidates and candidates[key].req.editable:
                candidate = candidates[key]
                # Create a new candidate with editable=False
                req = dataclasses.replace(candidate.req, editable=False)
                candidates[key] = candidate.copy_with(req)
        for key in requested:
            if key in candidates:
                candidates[key].requested = True
        return candidates

    def should_install_editables(self) -> bool:
        """Return whether to add editables"""
        if not self.install_self or "editables" in self.requested_candidates:
            return False
        # As editables may be added by the backend, we need to check the metadata
        try:
            metadata = self.self_candidate.prepare(self.environment).metadata
        except BuildError:
            return False
        return any(req.startswith("editables") for req in metadata.requires or [])

    @property
    def manager(self) -> InstallManager:
        if not self._manager:
            self._manager = self.get_manager(rename_pth=True)
        return self._manager

    def get_manager(self, rename_pth: bool = False) -> InstallManager:
        return self.environment.project.core.install_manager_class(
            self.environment, use_install_cache=self.use_install_cache, rename_pth=rename_pth
        )

    @property
    def self_key(self) -> str | None:
        if not self.install_self:
            return None
        name = self.environment.project.name
        if name:
            return normalize_name(name)
        return name

    def _should_update(self, dist: Distribution, can: Candidate) -> bool:
        """Check if the candidate should be updated"""
        backend = self.environment.project.backend
        if self.reinstall or can.req.editable:  # Always update if incoming is editable
            return True
        if is_editable(dist):  # only update editable if no_editable is True
            return bool(self.no_editable)
        if not can.req.is_named:
            dreq = Requirement.from_dist(dist)
            if not isinstance(dreq, FileRequirement):
                return True
            url = dreq.get_full_url()
            if dreq.is_local_dir:
                # We don't know whether a local dir has been changed, always update
                return True
            assert can.link is not None
            if url != backend.expand_line(can.link.url_without_fragment):
                return True
            direct_json = json.loads(content) if (content := dist.read_text("direct_url.json")) else None
            if not direct_json or "archive_info" not in direct_json:
                # We are not able to check, don't update
                return False
            dist_hash = direct_json["archive_info"]["hash"].replace("=", ":")
            return not any(dist_hash == file_hash["hash"] for file_hash in can.hashes)
        specifier = can.req.as_pinned_version(can.version).specifier
        return not specifier.contains(dist.version, prereleases=True)

    def compare_with_working_set(self) -> tuple[list[str], list[str], list[str]]:
        """Compares the candidates and return (to_add, to_update, to_remove)"""
        working_set = self.working_set
        candidates = self.candidates.copy()
        to_update: set[str] = set()
        to_remove: set[str] = set()
        to_add: set[str] = set()
        locked_repository = self.environment.project.get_locked_repository()
        all_candidate_keys = list(locked_repository.all_candidates)

        for key, dist in working_set.items():
            if key == self.self_key and self.install_self:
                continue
            if key in candidates:
                can = candidates.pop(key)
                if self._should_update(dist, can):
                    if working_set.is_owned(key):
                        to_update.add(key)
                    else:
                        to_add.add(key)
            elif (
                (self.only_keep or (self.clean and key not in all_candidate_keys))
                and key not in self.SEQUENTIAL_PACKAGES
                and working_set.is_owned(key)
            ):
                # Remove package only if it is not required by any group
                # Packages for packaging will never be removed
                to_remove.add(key)
        to_add.update(
            strip_extras(name)[0]
            for name, _ in candidates.items()
            if name != self.self_key and strip_extras(name)[0] not in working_set
        )
        return (sorted(to_add), sorted(to_update), sorted(to_remove))

    def synchronize(self) -> None:
        """Synchronize the working set with pinned candidates."""
        to_add, to_update, to_remove = self.compare_with_working_set()
        manager = self.get_manager()
        for key in to_add:
            can = self.candidates[key]
            termui.logger.info("Installing %s@%s...", key, can.version)
            manager.install(can)
        for key in to_update:
            can = self.candidates[key]
            dist = self.working_set[strip_extras(key)[0]]
            dist_version = dist.version
            termui.logger.info("Updating %s@%s -> %s...", key, dist_version, can.version)
            manager.overwrite(dist, can)
        for key in to_remove:
            dist = self.working_set[key]
            termui.logger.info("Removing %s@%s...", key, dist.version)
            manager.uninstall(dist)
        if self.install_self:
            self_key = self.self_key
            assert self_key
            word = "a" if self.no_editable else "an editable"
            termui.logger.info(f"Installing the project as {word} package...")
            if self_key in self.working_set:
                dist = self.working_set[strip_extras(self_key)[0]]
                manager.overwrite(dist, self.self_candidate)
            else:
                manager.install(self.self_candidate)
        termui.logger.info("Synchronization complete.")



from __future__ import annotations

import json
import os
import stat
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Iterator

from installer import install as _install
from installer._core import _process_WHEEL_file
from installer.destinations import SchemeDictionaryDestination, WheelDestination
from installer.exceptions import InvalidWheelSource
from installer.records import RecordEntry
from installer.sources import WheelContentElement, WheelSource
from installer.sources import WheelFile as _WheelFile

from pdm.models.cached_package import CachedPackage
from pdm.utils import make_file_executable

if TYPE_CHECKING:
    from typing import Any, BinaryIO, Iterable, Literal

    from installer.destinations import Scheme
    from installer.sources import WheelContentElement

    from pdm.environments import BaseEnvironment

    LinkMethod = Literal["symlink", "hardlink", "copy"]


class WheelFile(_WheelFile):
    @cached_property
    def dist_info_dir(self) -> str:
        namelist = self._zipfile.namelist()
        try:
            return next(name.split("/")[0] for name in namelist if name.split("/")[0].endswith(".dist-info"))
        except StopIteration:  # pragma: no cover
            canonical_name = super().dist_info_dir
            raise InvalidWheelSource(f"The wheel doesn't contain metadata {canonical_name!r}") from None


class PackageWheelSource(WheelSource):
    def __init__(self, package: CachedPackage) -> None:
        self.package = package

        distribution, version = package.path.name.split("-")[:2]
        super().__init__(distribution, version)

    @cached_property
    def dist_info_dir(self) -> str:
        return self.package.dist_info.name

    @property
    def dist_info_filenames(self) -> list[str]:
        return os.listdir(self.package.dist_info)

    def read_dist_info(self, filename: str) -> str:
        return self.package.dist_info.joinpath(filename).read_text("utf-8")

    def iter_files(self) -> Iterable[Path]:
        for root, _, files in os.walk(self.package.path):
            for file in files:
                if Path(root) == self.package.path and file in CachedPackage.cache_files:
                    continue
                yield Path(root, file)

    def get_contents(self) -> Iterator[WheelContentElement]:
        from installer.records import parse_record_file

        record_lines = self.read_dist_info("RECORD").splitlines()
        records = parse_record_file(record_lines)
        record_mapping = {record[0]: record for record in records}

        for item in self.iter_files():
            fn = item.relative_to(self.package.path).as_posix()

            # Pop record with empty default, because validation is handled by `validate_record`
            record = record_mapping.pop(fn, (fn, "", ""))

            # Borrowed from:
            # https://github.com/pypa/pip/blob/0f21fb92/src/pip/_internal/utils/unpacking.py#L96-L100
            mode = item.stat().st_mode
            is_executable = bool(mode and stat.S_ISREG(mode) and mode & 0o111)

            with item.open("rb") as stream:
                yield record, stream, is_executable


class InstallDestination(SchemeDictionaryDestination):
    def __init__(
        self,
        *args: Any,
        link_method: LinkMethod = "copy",
        rename_pth: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.link_method = link_method
        self.rename_pth = rename_pth

    def _compile_bytecode(self, scheme: Scheme, record: RecordEntry) -> None:
        if self.link_method == "symlink":
            return
        super()._compile_bytecode(scheme, record)

    def finalize_installation(
        self,
        scheme: Scheme,
        record_file_path: str,
        records: Iterable[tuple[Scheme, RecordEntry]],
    ) -> None:
        if os.name != "nt":
            return super().finalize_installation(scheme, record_file_path, records)

        from installer.destinations import construct_record_file

        def prefix_for_scheme(file_scheme: Scheme) -> str | None:
            if file_scheme == scheme:
                return None
            try:
                path = os.path.relpath(self.scheme_dict[file_scheme], start=self.scheme_dict[scheme])
            except ValueError:
                path = os.path.abspath(self.scheme_dict[file_scheme])
            return path + "/"

        record_list = list(records)
        with construct_record_file(record_list, prefix_for_scheme) as record_stream:
            self.write_to_fs(scheme, record_file_path, record_stream, is_executable=False)

        for file_scheme, record in record_list:
            self._compile_bytecode(file_scheme, record)

    def write_to_fs(self, scheme: Scheme, path: str, stream: BinaryIO, is_executable: bool) -> RecordEntry:
        from installer.records import Hash
        from installer.utils import copyfileobj_with_hashing

        target_path = os.path.join(self.scheme_dict[scheme], path)
        if os.path.exists(target_path):
            os.unlink(target_path)

        os.makedirs(os.path.dirname(target_path), exist_ok=True)

        if self.rename_pth and target_path.endswith(".pth") and "/" not in path:
            # Postpone the creation of pth files since it may cause race condition
            # when multiple packages are installed at the same time.
            target_path += ".pdmtmp"
        if self.link_method == "copy" or not hasattr(stream, "name"):
            with open(target_path, "wb") as f:
                hash_, size = copyfileobj_with_hashing(stream, f, self.hash_algorithm)
        else:
            src_path = stream.name
            # create links, we don't need the stream anymore
            stream.close()
            if self.link_method == "symlink":
                os.symlink(src_path, target_path)
            else:  # hardlink
                os.link(src_path, target_path)
            hash_ = ""
            size = os.path.getsize(target_path)

        if is_executable:
            make_file_executable(target_path)
        return RecordEntry(path, Hash(self.hash_algorithm, hash_), size)


def _get_link_method(cache_method: str) -> LinkMethod:
    from pdm import utils

    if "symlink" in cache_method and utils.fs_supports_link_method("symlink"):
        return "symlink"

    if "link" in cache_method and utils.fs_supports_link_method("link"):
        return "hardlink"
    return "copy"


def install_wheel(
    wheel: Path,
    environment: BaseEnvironment,
    direct_url: dict[str, Any] | None = None,
    install_links: bool = False,
    rename_pth: bool = False,
    requested: bool = False,
) -> str:
    """Only create .pth files referring to the cached package.
    If the cache doesn't exist, create one.
    """
    interpreter = str(environment.interpreter.executable)
    script_kind = environment.script_kind
    # the cache_method can be any of "symlink", "hardlink", "copy" and "pth"
    cache_method: str = environment.project.config["install.cache_method"]
    dist_name = wheel.name.split("-")[0]
    link_method: LinkMethod | None
    if not install_links or dist_name == "editables":
        link_method = "copy"
    else:
        link_method = _get_link_method(cache_method)

    additional_metadata: dict[str, bytes] = {"INSTALLER": b"pdm"}
    if direct_url is not None:
        additional_metadata["direct_url.json"] = json.dumps(direct_url, indent=2).encode()
    if requested:
        additional_metadata["REQUESTED"] = b""

    destination = InstallDestination(
        scheme_dict=environment.get_paths(dist_name),
        interpreter=interpreter,
        script_kind=script_kind,
        link_method=link_method,
        rename_pth=rename_pth,
    )
    if install_links:
        package = environment.project.package_cache.cache_wheel(wheel)
        source = PackageWheelSource(package)
        if link_method == "symlink":
            # Track usage when symlink is used
            additional_metadata["REFER_TO"] = package.path.as_posix().encode()
        dist_info_dir = install(source, destination=destination, additional_metadata=additional_metadata)
        if link_method == "symlink":
            package.add_referrer(dist_info_dir)
    else:
        with WheelFile.open(wheel) as source:
            dist_info_dir = install(source, destination=destination, additional_metadata=additional_metadata)
    return dist_info_dir


def install(
    source: WheelSource, destination: WheelDestination, additional_metadata: dict[str, bytes] | None = None
) -> str:
    """A lower level installation method that is copied from installer
    but is controlled by extra parameters.

    Return the .dist-info path
    """
    _install(source, destination, additional_metadata=additional_metadata or {})
    root_scheme = _process_WHEEL_file(source)
    return os.path.join(destination.scheme_dict[root_scheme], source.dist_info_dir)



import argparse
import re
import sys
from typing import cast

from pdm import termui
from pdm.cli import actions
from pdm.cli.commands.base import BaseCommand
from pdm.cli.filters import GroupSelection
from pdm.cli.hooks import HookManager
from pdm.cli.options import (
    config_setting_option,
    groups_group,
    lock_strategy_group,
    lockfile_option,
    no_isolation_option,
    override_option,
    skip_option,
)
from pdm.models.markers import EnvSpec
from pdm.models.specifiers import PySpecSet
from pdm.project import Project
from pdm.utils import convert_to_datetime


class Command(BaseCommand):
    """Resolve and lock dependencies"""

    arguments = (
        *BaseCommand.arguments,
        lockfile_option,
        no_isolation_option,
        config_setting_option,
        override_option,
        skip_option,
        groups_group,
        lock_strategy_group,
    )

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--refresh",
            action="store_true",
            help="Refresh the content hash and file hashes in the lock file",
        )

        parser.add_argument(
            "--check",
            action="store_true",
            help="Check if the lock file is up to date and quit",
        )
        parser.add_argument(
            "--update-reuse",
            action="store_const",
            dest="update_strategy",
            default="all",
            const="reuse",
            help="Reuse pinned versions already present in lock file if possible",
        )
        parser.add_argument(
            "--update-reuse-installed",
            action="store_const",
            dest="update_strategy",
            const="reuse-installed",
            help="Reuse installed packages if possible",
        )
        parser.add_argument(
            "--exclude-newer",
            help="Exclude packages newer than the given UTC date in format `YYYY-MM-DD[THH:MM:SSZ]` "
            "or relative time `N{d|h|w}`",
            type=convert_to_datetime,
        )

        target_group = parser.add_argument_group("Lock Target")
        target_group.add_argument("--python", help="The Python range to lock for. E.g. `>=3.9`, `==3.12.*`")
        target_group.add_argument(
            "--platform",
            help="The platform to lock for. E.g. `windows`, `linux`, `macos`, `manylinux_2_17_x86_64`. "
            "See docs for available choices: http://pdm-project.org/en/latest/usage/lock-targets/",
        )
        target_group.add_argument(
            "--implementation",
            help="The Python implementation to lock for. E.g. `cpython`, `pypy`, `pyston`",
        )
        target_group.add_argument("--append", action="store_true", help="Append the result to the current lock file")

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        if options.check:
            strategy = actions.check_lockfile(project, False)
            if strategy:
                project.core.ui.echo(
                    f"[error]{termui.Emoji.FAIL}[/] Lockfile is [error]out of date[/].",
                    err=True,
                    verbosity=termui.Verbosity.DETAIL,
                )
                sys.exit(1)
            else:
                project.core.ui.echo(
                    f"[success]{termui.Emoji.SUCC}[/] Lockfile is [success]up to date[/].",
                    err=True,
                    verbosity=termui.Verbosity.DETAIL,
                )
                sys.exit(0)
        selection = GroupSelection.from_options(project, options)
        strategy = options.update_strategy
        if options.exclude_newer:
            # cli args override pyproject config if present
            project.core.state.exclude_newer = options.exclude_newer
        if project.core.state.exclude_newer:
            strategy = "all"
            if strategy != options.update_strategy:
                project.core.ui.info("--exclude-newer is set, forcing --update-all")
        env_spec: EnvSpec | None = None
        if any([options.python, options.platform, options.implementation]):
            replace_dict = {}
            if options.python:
                if re.match(r"[\d.]+", options.python):
                    options.python = f">={options.python}"
                replace_dict["requires_python"] = PySpecSet(options.python)
            if options.platform:
                replace_dict["platform"] = options.platform
            if options.implementation:
                replace_dict["implementation"] = options.implementation
            env_spec = project.environment.allow_all_spec.replace(**replace_dict)

        actions.do_lock(
            project,
            refresh=options.refresh,
            strategy=cast(str, strategy),
            groups=selection.all(),
            strategy_change=options.strategy_change,
            hooks=HookManager(project, options.skip),
            env_spec=env_spec,
            append=options.append,
        )



from __future__ import annotations

import os

from pdm.builders.base import EnvBuilder, wrap_error


class WheelBuilder(EnvBuilder):
    """Build wheel in isolated env with managed Python."""

    @wrap_error
    def prepare_metadata(self, out_dir: str) -> str:
        if self.isolated:
            self.install(self._requires, shared=True)
            requires = self._hook.get_requires_for_build_wheel(self.config_settings)
            self.install(requires)
        filename = self._hook.prepare_metadata_for_build_wheel(out_dir, self.config_settings)
        return os.path.join(out_dir, filename)

    @wrap_error
    def build(self, out_dir: str, metadata_directory: str | None = None) -> str:
        if self.isolated:
            self.install(self._requires, shared=True)
            requires = self._hook.get_requires_for_build_wheel(self.config_settings)
            self.install(requires)
        filename = self._hook.build_wheel(out_dir, self.config_settings, metadata_directory)
        return os.path.join(out_dir, filename)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from pdm.exceptions import CandidateInfoNotFound, CandidateNotFound
from pdm.models.candidates import Candidate
from pdm.models.repositories.base import BaseRepository, CandidateMetadata, cache_result
from pdm.models.requirements import Requirement, filter_requirements_with_extras
from pdm.models.search import SearchResultParser

if TYPE_CHECKING:
    from typing import Callable, Iterable

    from pdm._types import SearchResults


class PyPIRepository(BaseRepository):
    """Get package and metadata from PyPI source."""

    DEFAULT_INDEX_URL = "https://pypi.org"

    @cache_result
    def _get_dependencies_from_json(self, candidate: Candidate) -> CandidateMetadata:  # pragma: no cover
        if not candidate.name or not candidate.version:
            # Only look for json api for named requirements.
            raise CandidateInfoNotFound(candidate)
        sources = self.get_filtered_sources(candidate.req)
        url_prefixes = [
            proc_url[:-7]  # Strip "/simple".
            for proc_url in (raw_url.rstrip("/") for raw_url in (source.url for source in sources) if raw_url)
            if proc_url.endswith("/simple")
        ]
        session = self.environment.session
        for prefix in url_prefixes:
            json_url = f"{prefix}/pypi/{candidate.name}/{candidate.version}/json"
            resp = session.get(json_url)
            if resp.is_error:
                continue

            info = resp.json()["info"]

            requires_python = info["requires_python"] or ""
            summary = info["summary"] or ""
            try:
                requirement_lines = info["requires_dist"] or []
            except KeyError:
                requirement_lines = info["requires"] or []
            requirements = filter_requirements_with_extras(requirement_lines, candidate.req.extras or ())
            return CandidateMetadata(requirements, requires_python, summary)
        raise CandidateInfoNotFound(candidate)

    def dependency_generators(self) -> Iterable[Callable[[Candidate], CandidateMetadata]]:
        yield self._get_dependencies_from_cache
        if self.environment.project.config["pypi.json_api"]:
            yield self._get_dependencies_from_json
        yield self._get_dependencies_from_metadata

    def _find_candidates(self, requirement: Requirement, minimal_version: bool) -> Iterable[Candidate]:
        from unearth.utils import LazySequence

        sources = self.get_filtered_sources(requirement)
        req_name = cast(str, requirement.project_name)
        with self.environment.get_finder(sources, env_spec=self.env_spec, minimal_version=minimal_version) as finder:
            cans = LazySequence(
                Candidate.from_installation_candidate(c, requirement)
                for c in finder.find_all_packages(req_name, allow_yanked=requirement.is_pinned)
            )
        if not cans:
            raise CandidateNotFound(
                f"Unable to find candidates for {req_name}. There may "
                "exist some issues with the package name or network condition."
            )
        return cans

    def search(self, query: str) -> SearchResults:
        pypi_simple = self.sources[0].url.rstrip("/")  # type: ignore[union-attr]

        if pypi_simple.endswith("/simple"):
            search_url = pypi_simple[:-6] + "search"
        else:
            search_url = pypi_simple + "/search"

        session = self.environment.session
        resp = session.get(search_url, params={"q": query})
        if resp.status_code == 404:  # pragma: no cover
            self.environment.project.core.ui.warn(
                f"{pypi_simple!r} doesn't support '/search' endpoint, fallback "
                f"to {self.DEFAULT_INDEX_URL!r} now.\n"
                "This may take longer depending on your network condition.",
            )
            resp = session.get(f"{self.DEFAULT_INDEX_URL}/search/", params={"q": query}, follow_redirects=True)
        parser = SearchResultParser()
        resp.raise_for_status()
        parser.feed(resp.text)
        return parser.results



from __future__ import annotations

import tempfile
from collections.abc import Iterator
from contextlib import ExitStack, contextmanager
from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from typing import Any, cast

import tomlkit

from pdm.models.candidates import Candidate
from pdm.models.markers import Marker, get_marker
from pdm.models.repositories import LockedRepository, Package
from pdm.models.requirements import FileRequirement, Requirement, VcsRequirement, parse_requirement, strip_extras
from pdm.project.core import Project
from pdm.utils import get_requirement_from_override, normalize_name


@dataclass
class _UvFileBuilder:
    project: Project
    requires_python: str
    requirements: list[Requirement]
    locked_repository: LockedRepository
    stack: ExitStack = field(default_factory=ExitStack, init=False)

    @cached_property
    def default_source(self) -> str:
        return cast(str, self.project.sources[0].url)

    def __post_init__(self) -> None:
        self._enter_path(self.project.root / "uv.lock")

    def build_pyproject_toml(self) -> Path:
        data = self.project.pyproject.open_for_read()

        uv_overrides = []
        for override_key, override_value in (
            data.get("tool", {}).get("pdm", {}).get("resolution", {}).get("overrides", {}).items()
        ):
            uv_overrides.append(f"{get_requirement_from_override(override_key, override_value)}")

        if uv_overrides:
            data.setdefault("tool", {}).setdefault("uv", {}).setdefault("override-dependencies", []).extend(
                uv_overrides
            )

        data.setdefault("project", {})["requires-python"] = self.requires_python
        data.pop("dependency-groups", None)
        data.setdefault("project", {}).pop("optional-dependencies", None)
        sources = {}
        collected_deps: dict[str, list[str]] = {}
        for dep in self.requirements:
            if isinstance(dep, FileRequirement):
                entry = self._get_name(dep)
                sources[entry] = self._build_source(dep)
            else:
                entry = dep.as_line()
            for group in dep.groups:
                collected = collected_deps.setdefault(group, [])
                if entry not in collected:
                    collected.append(entry)

        for group, deps in collected_deps.items():
            if group == "default":
                data.setdefault("project", {})["dependencies"] = deps
            else:
                data.setdefault("project", {}).setdefault("optional-dependencies", {})[group] = deps

        if sources:
            data.setdefault("tool", {}).setdefault("uv", {}).setdefault("sources", {}).update(sources)

        path = self._enter_path(self.project.root / "pyproject.toml")
        with path.open("w", newline="", encoding="utf-8") as f:
            tomlkit.dump(data, f)
        return path

    def _enter_path(self, path: Path) -> Path:
        if path.exists():
            name = tempfile.mktemp(dir=path.parent, prefix=f"{path.name}.")
            backup = path.rename(name)

            @self.stack.callback
            def restore() -> None:
                path.unlink(True)
                backup.rename(path)
        else:
            self.stack.callback(path.unlink, True)
        return path

    def build_uv_lock(self, include_self: bool = False) -> Path:
        locked_repo = self.locked_repository
        packages: list[dict[str, Any]] = []
        for key in locked_repo.packages:
            if "[" in key[0]:  # skip entries with extras
                continue
            # Merge related entries with the same name and version
            related_packages = [
                p for k, p in locked_repo.packages.items() if strip_extras(k[0])[0] == key[0] and k[1:] == key[1:]
            ]
            packages.append(self._build_lock_entry(related_packages))
        if name := self.project.name:
            version = self.project.pyproject.metadata.get("version", "0.0.0")
            this_package = {
                "name": normalize_name(name),
                "version": version,
                "source": {"editable" if include_self else "virtual": "."},
            }
            dependencies: list[dict[str, Any]] = []
            optional_dependencies: dict[str, list[dict[str, Any]]] = {}
            for req in self.requirements:
                if (dep := self._make_dependency(None, req)) is None:
                    continue
                for group in req.groups:
                    if group == "default":
                        target_group = dependencies
                    else:
                        target_group = optional_dependencies.setdefault(group, [])
                    if dep not in target_group:
                        target_group.append(dep)
            if dependencies:
                this_package["dependencies"] = dependencies  # type: ignore[assignment]
            if optional_dependencies:
                this_package["optional-dependencies"] = optional_dependencies
            packages.append(this_package)

        data = {"version": 1, "requires-python": self.requires_python}
        if packages:
            data["package"] = packages
        path = self.project.root / "uv.lock"
        with path.open("w", newline="", encoding="utf-8") as f:
            tomlkit.dump(data, f)
        return path

    def _get_name(self, req: FileRequirement) -> str:
        if req.key:
            return req.key
        can = Candidate(req).prepare(self.project.environment)
        return normalize_name(can.metadata.name)

    def _build_source(self, req: FileRequirement) -> dict[str, Any]:
        result: dict[str, Any]
        if isinstance(req, VcsRequirement):
            result = {req.vcs: req.repo}
            if req.ref:
                result["rev"] = req.ref
        elif req.path:
            result = {"path": req.str_path}
        else:
            result = {"url": req.url}
        if req.editable:
            result["editable"] = True
        return result

    def _build_lock_source(self, req: Requirement) -> dict[str, Any]:
        if isinstance(req, VcsRequirement):
            return {req.vcs: f"{req.repo}?rev={req.ref}#{req.revision}"}
        elif isinstance(req, FileRequirement):
            if req.editable:
                return {"editable": req.str_path}
            else:
                return {"url": req.url}
        else:
            return {"registry": self.default_source}

    def _build_lock_entry(self, packages: list[Package]) -> dict[str, Any]:
        packages.sort(key=lambda x: len(x.candidate.req.extras or []))
        candidate = packages[0].candidate
        req = candidate.req
        result: dict[str, Any] = {
            "name": candidate.name,
            "version": candidate.version,
            "source": self._build_lock_source(req),
        }
        for file_hash in candidate.hashes:
            filename = file_hash.get("url", file_hash.get("file", ""))
            is_wheel = filename.endswith(".whl")
            item = {"url": file_hash.get("url", filename), "hash": file_hash["hash"]}
            if is_wheel:
                result.setdefault("wheels", []).append(item)
            else:
                result["sdist"] = item
        optional_dependencies: dict[str, list[dict[str, Any]]] = {}
        for package in packages:
            if package.dependencies is None:
                continue
            if not package.candidate.req.extras:
                deps = [
                    self._make_dependency(package.candidate, parse_requirement(dep)) for dep in package.dependencies
                ]
                result["dependencies"] = [dep for dep in deps if dep is not None]
            else:
                deps = [
                    self._make_dependency(package.candidate, parse_requirement(dep))
                    for dep in package.dependencies
                    if parse_requirement(dep).key != candidate.req.key
                ]
                deps = [dep for dep in deps if dep is not None]
                for extra in package.candidate.req.extras:
                    # XXX: when depending on a package with extras, the extra dependencies are encoded in
                    # the corresponding group under optional-dependencies. But in case multiple extras are requested,
                    # the same dependencies get duplicated in those groups, but it's okay if each single extra is
                    # never requested alone.
                    if extra not in optional_dependencies:
                        optional_dependencies[extra] = deps  # type: ignore[assignment]

        if optional_dependencies:
            result["optional-dependencies"] = optional_dependencies
        return result

    def _make_dependency(self, parent: Candidate | None, req: Requirement) -> dict[str, Any] | None:
        locked_repo = self.locked_repository
        parent_marker = req.marker or get_marker("")
        if parent is not None:
            parent_marker &= parent.req.marker or get_marker("")
        matching_entries = [e for k, e in locked_repo.packages.items() if k[0] == req.key]

        def marker_match(marker: Marker | None) -> bool:
            return not (parent_marker & (marker or get_marker(""))).is_empty()

        if not matching_entries:
            return None
        result: dict[str, Any] = {}
        if len(matching_entries) == 1:
            candidate = matching_entries[0].candidate
            multiple = False
        else:
            candidate = next(e.candidate for e in matching_entries if marker_match(e.candidate.req.marker))
            multiple = True
        result.update({"name": candidate.name})
        if multiple:
            result.update(version=candidate.version, source=self._build_lock_source(candidate.req))
        if req.extras:
            result["extra"] = list(req.extras)
        if req.marker is not None:
            result["marker"] = str(req.marker)
        return result


@contextmanager
def uv_file_builder(
    project: Project, requires_python: str, requirements: list[Requirement], locked_repository: LockedRepository
) -> Iterator[_UvFileBuilder]:
    builder = _UvFileBuilder(project, requires_python, requirements, locked_repository)
    with builder.stack:
        yield builder
